import data.fintype

inductive B : Type
| O : B
| I : B

export B (O I)

namespace B

protected def repr : B → string
| O := "O"
| I := "I"

instance : has_repr B :=
⟨B.repr⟩

lemma O_ne_I (H : O = I) : false := 
by contradiction

instance : decidable_eq B
| O O := is_true rfl
| O I := is_false O_ne_I
| I O := is_false (ne.symm O_ne_I)
| I I := is_true rfl

open fintype 

instance : fintype B := 
begin
have h : ∀ x : B, x ∈ [O,I], 
    {intro x, cases x, 
        {left, refl},
        {right, left, refl}},
apply fintype.of_list _ h
end

@[simp] lemma fintype.card_B : fintype.card B = 2 := rfl

instance decidable_eq_vectorB {n : ℕ} : decidable_eq (vector B n) :=
by apply_instance

instance fintype_vectorB {n : ℕ} : fintype (vector B n) := 
vector.fintype

@[simp] lemma card_vectorB (n : ℕ) : 
    card (vector B n) = 2 ^ n :=
by {rw card_vector, rw fintype.card_B}

end B
