import data.vector data.fintype
import .B .InsDel .dS .NumOsNumIs

variable {α : Type}
variable [decidable_eq α] 

open nat list vector fintype finset

namespace set 

def is_DelCode {n : ℕ} (C : set (vector α n)) := 
  ∀ X Y ∈ C, X ≠ Y → dS(X) ∩ dS(Y) = ∅

def is_DelCode' {n : ℕ} (C : set (vector α n)) := 
  ∀ X Y ∈ C, (∃ i j : ℕ, sDel X i = sDel Y j) → X = Y 

def is_DelCode'' {n : ℕ} (C : set (vector α n)) := 
  ∀ X Y ∈ C, X ≠ Y → ∀ i j : ℕ, sDel X i ≠ sDel Y j

lemma DelCode'_of_DelCode'' 
  {n : ℕ} (C : set (vector α n)) (H : is_DelCode'' C) :
  is_DelCode' C :=
begin
intros X Y Xin Yin h,
cases h with i h_i, cases h_i with j h_ij, 
cases decidable.em (X = Y) with heq hneq,
  {apply heq},
  {apply absurd h_ij,
   apply H X Y Xin Yin hneq}
end

lemma DelCode''_of_DelCode' 
  {n : ℕ} (C : set (vector α n)) (H : is_DelCode' C) :
  is_DelCode'' C :=
begin
intros X Y Xin Yin ne i j,
cases decidable.em (sDel X i = sDel Y j) with heq hneq,
  {have h : X = Y,
    {apply H X Y Xin Yin,
     apply exists.intro i, apply exists.intro j, apply heq},
   contradiction},
  {apply hneq}
end

lemma DelCode''_iff_DelCode' {n : ℕ} (C : set (vector α n)) :
    is_DelCode'' C ↔ is_DelCode' C :=
begin
apply iff.intro,
  {apply DelCode'_of_DelCode''},
  {apply DelCode''_of_DelCode'}
end

lemma DelCode_of_DelCode'
  {n : ℕ} (C : set (vector α n)) (H : is_DelCode' C) :
  is_DelCode C :=
begin
intros X Y Xin Yin ne, 
apply set.subset.antisymm,
  {intros Z inZ, cases inZ with inZl inZr, 
   rw dS_def at inZl, cases inZl with i inZli, 
   rw dS_def at inZr, cases inZr with j inZrj,  
   have h1 : ∃ i j : ℕ, sDel X i = sDel Y j,
    {apply exists.intro i, apply exists.intro j, rw ← inZli, rw ← inZrj},
   have h2 : X = Y, apply (H X Y Xin Yin h1),
   rw ne_from_not_eq at ne, contradiction},
  {apply set.empty_subset}
end

lemma DelCode'_of_DelCode
  {n : ℕ} (C : set (vector α n)) (H : is_DelCode C) :
  is_DelCode' C :=
begin
rw ← DelCode''_iff_DelCode', 
intros X Y inX inY ne i j, 
cases decidable.em (sDel X i = sDel Y j) with heq hneq,
  {have h1 : dS X ∩ dS Y = ∅,
    {apply H X Y inX inY, apply ne},
   have h2 : dS X ∩ dS Y ≠ ∅,
    {rw set.ne_empty_iff_exists_mem, 
     apply exists.intro (sDel X i), apply and.intro,
      {apply exists.intro i, refl},
      {apply exists.intro j, rw ← heq}},
   contradiction},
  {apply hneq}
end

lemma DelCode'_iff_DelCode
  {n : ℕ} (C : set (vector α n)) :
  is_DelCode' C ↔ is_DelCode C:=
begin
apply iff.intro,
  {apply DelCode_of_DelCode'},
  {apply DelCode'_of_DelCode}
end

end set 

namespace finset

variable [fintype α]

def is_DelCode {n : ℕ} (C : finset (vector α n)) := 
  ∀ X Y ∈ C, X ≠ Y → dS(X) ∩ dS(Y) = ∅

def is_DelCode' {n : ℕ} (C : finset (vector α n)) := 
  ∀ X Y ∈ C, (∃ i j : ℕ, sDel X i = sDel Y j) → X = Y 

def is_DelCode'' {n : ℕ} (C : finset (vector α n)) := 
  ∀ X Y ∈ C, X ≠ Y → ∀ i j : ℕ, sDel X i ≠ sDel Y j

lemma DelCode'_of_DelCode'' 
  {n : ℕ} (C : finset (vector α n)) (H : is_DelCode'' C) :
  is_DelCode' C :=
begin
intros X Y Xin Yin h,
cases h with i h_i, cases h_i with j h_ij, 
cases decidable.em (X = Y) with heq hneq,
  {apply heq},
  {apply absurd h_ij,
   apply H X Y Xin Yin hneq}
end

lemma DelCode''_of_DelCode' 
  {n : ℕ} (C : finset (vector α n)) (H : is_DelCode' C) :
  is_DelCode'' C :=
begin
intros X Y Xin Yin ne i j,
cases decidable.em (sDel X i = sDel Y j) with heq hneq,
  {have h : X = Y,
    {apply H X Y Xin Yin,
     apply exists.intro i, apply exists.intro j, apply heq},
   contradiction},
  {apply hneq}
end

lemma DelCode''_iff_DelCode' {n : ℕ} (C : finset (vector α n)) :
    is_DelCode'' C ↔ is_DelCode' C :=
begin
apply iff.intro,
  {apply DelCode'_of_DelCode''},
  {apply DelCode''_of_DelCode'}
end

lemma DelCode_of_DelCode'
  {n : ℕ} (C : finset (vector α n)) (H : is_DelCode' C) :
  is_DelCode C :=
begin
intros X Y Xin Yin ne, 
apply set.subset.antisymm,
  {intros Z inZ, cases inZ with inZl inZr, 
   rw dS_def at inZl, cases inZl with i inZli, 
   rw dS_def at inZr, cases inZr with j inZrj,  
   have h1 : ∃ i j : ℕ, sDel X i = sDel Y j,
    {apply exists.intro i, apply exists.intro j, rw ← inZli, rw ← inZrj},
   have h2 : X = Y, apply (H X Y Xin Yin h1),
   rw ne_from_not_eq at ne, contradiction},
  {apply set.empty_subset}
end

lemma DelCode'_of_DelCode
  {n : ℕ} (C : finset (vector α n)) (H : is_DelCode C) :
  is_DelCode' C :=
begin
rw ← DelCode''_iff_DelCode', 
intros X Y inX inY ne i j, 
cases decidable.em (sDel X i = sDel Y j) with heq hneq,
  {have h1 : dS X ∩ dS Y = ∅,
    {apply H X Y inX inY, apply ne},
   have h2 : dS X ∩ dS Y ≠ ∅,
    {rw set.ne_empty_iff_nonempty, 
     apply exists.intro (sDel X i), apply and.intro,
      {apply exists.intro i, refl},
      {apply exists.intro j, rw ← heq}},
   contradiction},
  {apply hneq}
end

lemma DelCode'_iff_DelCode
  {n : ℕ} (C : finset (vector α n)) :
  is_DelCode' C ↔ is_DelCode C:=
begin
apply iff.intro,
  {apply DelCode_of_DelCode'},
  {apply DelCode'_of_DelCode}
end

def dis_DelCode {n : ℕ} (C : finset (vector α n)) := 
  ∀ X ∈ C, ∀ Y ∈ C, (∃ i ∈ range (n + 1), ∃ j ∈ range (n + 1), sDel X i = sDel Y j) → X = Y

lemma DelCode_of_dDelCode 
  {n : ℕ} (C : finset (vector α n)) (H : dis_DelCode C) :
  is_DelCode C :=
begin
rw ← DelCode'_iff_DelCode,
intros X Y Xin Yin h, 
cases h with i h_i,  
cases exists_sDel_range X i with i' h_i',
cases h_i' with iin h_i',
cases h_i with j h_j,  
cases exists_sDel_range Y j with j' h_j', 
cases h_j' with jin h_j',
apply H X Xin Y Yin,
apply exists.intro i', apply exists.intro iin,
apply exists.intro j', apply exists.intro jin,
rw ← h_i', rw ← h_j', apply h_j
end

lemma dDelCode_of_DelCode 
  {n : ℕ} (C : finset (vector α n)) (H : is_DelCode C) :
  dis_DelCode C:=
begin
intros X Xin Y Yin h,
cases h with i h_i, cases h_i with iin h_i, 
cases h_i with j h_ij, cases h_ij with jin h_ij, 
rw ← DelCode'_iff_DelCode at H,
apply H X Y Xin Yin,
apply exists.intro i, apply exists.intro j, apply h_ij
end

lemma dDelCode_iff_DelCode {n : ℕ} (C : finset (vector α n)) :
  dis_DelCode C ↔ is_DelCode C :=
begin
apply iff.intro,
  {apply DelCode_of_dDelCode},
  {apply dDelCode_of_DelCode}
end

instance decidable_pred_dis_DelCode {n : ℕ} : 
  decidable_pred (@dis_DelCode α _inst_1 _inst_2 n) :=
by {intro s, rw dis_DelCode, apply_instance}

instance decidable_pred_is_DelCode {n : ℕ} : 
  decidable_pred (@is_DelCode α _inst_1 _inst_2 n) :=
begin
intro s, apply decidable_of_iff (dis_DelCode s), 
apply (dDelCode_iff_DelCode s)
end

def DelCodes {n : ℕ} : finset (finset (vector B n)) := 
  filter is_DelCode (powerset univ) 

lemma DelCodes_def {n : ℕ} (C : finset (vector B n)) : 
  C ∈ @DelCodes n ↔ is_DelCode C :=
begin
apply iff.intro,
  {intros H, unfold DelCodes at H, rw mem_filter at H, apply H.right},
  {intros H, unfold DelCodes, rw mem_filter, apply and.intro,
    {rw mem_powerset, apply subset_univ},
    {apply H}}
end

lemma DelCode_filter
  {n : ℕ} (C : finset (vector B n)) (H : is_DelCode C)
  (p : vector B n → Prop) [decidable_pred p] :
  is_DelCode (filter p C) :=
begin
intros x y xin yin ne,
rw mem_filter at xin, rw mem_filter at yin, 
apply H x y xin.left yin.left ne
end

lemma DelCode_insert
  {n : ℕ} (C : finset (vector B n)) (H1 : is_DelCode C)
  (x : vector B n) (H2 : ∀ y ∈ C, dS(x) ∩ dS(y) = ∅):
  is_DelCode (insert x C) :=
begin
intros y z yin zin ne,
rw mem_insert at yin, rw mem_insert at zin, 
cases yin with yinl yinr,
  {cases zin with zinl zinr, 
    {rw ← zinl at yinl, contradiction},
    {rw yinl, apply H2 z zinr}},
  {cases zin with zinl zinr, 
    {rw zinl, rw set.inter_comm, apply H2 y yinr},
    {apply H1 y z yinr zinr ne}}
end


def not (p : α → Prop) : α → Prop :=
  λ x, ¬ p x

instance decidable_pred_not (p : α → Prop) [decidable_pred p] : 
  decidable_pred (not p) :=
by {intro s, unfold not, apply_instance}

lemma not_not (p : α → Prop) [decidable_pred p] :
  not (not p) = p :=
begin
rw function.funext_iff, 
intro a, unfold not, rw not_not
end 

lemma filter_union_not
  (C : finset α) (p : α → Prop) [decidable_pred p] :
  filter p C ∪ filter (not p) C = C :=
begin
rw subset.antisymm_iff, apply and.intro,
  {intros x h, rw mem_union at h, cases h with hl hr,
    {rw mem_filter at hl, apply hl.left},
    {rw mem_filter at hr, apply hr.left}},
  {intros x h, rw mem_union,
   cases decidable.em (p x) with hp hnp,
    {left, rw mem_filter, apply and.intro h hp},
    {right, rw mem_filter, apply and.intro h hnp}}
end

lemma filter_inter_not 
  (C : finset α) (p : α → Prop) [decidable_pred p] :
  filter p C ∩ filter (not p) C = ∅ :=
begin
rw subset.antisymm_iff, apply and.intro,
  {intros x h, rw mem_inter at h, cases h with hl hr,
   rw mem_filter at hl, rw mem_filter at hr, 
   apply absurd hl.right hr.right},
  {apply empty_subset}
end 

lemma card_filter_add_filter
  (C : finset α) (p : α → Prop) [decidable_pred p] :
  card C = card (filter p C) + card (filter (not p) C) :=
begin
rw ← card_union_add_card_inter,
rw filter_union_not, rw filter_inter_not,
rw card_empty, rw nat.add_zero
end

lemma card_filter
  (C : finset α) (p : α → Prop) [decidable_pred p] :
  card (filter p C) = card C - card (filter (not p) C) :=
by {rw card_filter_add_filter C p, rw nat.add_sub_cancel}

lemma card_filter_not
  (C : finset α) (p : α → Prop) [decidable_pred p] :
  card (filter (not p) C) = card C - card (filter p C) :=
by {rw card_filter_add_filter C p, rw nat.add_sub_cancel_left}

def wt_le {n : ℕ} (x : vector B n) : Prop :=
  wt x ≤ 1

instance decidable_pred_wt_le {n : ℕ} : 
  decidable_pred (@wt_le n) :=
by {intro s, unfold wt_le, apply_instance}

lemma card_filter_wt_le_one
  {n : ℕ} (C : finset (vector B n)) (H : is_DelCode C):
  card (filter wt_le C) ≤ 1 :=
begin
cases decidable.em(filter wt_le C = ∅) with hem hnem,
  {rw hem, rw card_empty, apply nat.zero_le},
  {cases exists_mem_of_ne_empty hnem with a ain,
   apply le_of_eq, rw card_eq_one, apply exists.intro a,
   rw eq_singleton_iff_unique_mem, apply and.intro,
    {apply ain},
    {intros x xin, 
     rw mem_filter at xin, rw mem_filter at ain,
     apply DelCode'_of_DelCode C H,
      {apply xin.left},
      {apply ain.left},
      {cases sDel_of_le_wt1 x xin.right with i xrepO,
       cases sDel_of_le_wt1 a ain.right with j arepO,
       apply exists.intro i, apply exists.intro j,
       rw xrepO, rw arepO}}}
end

def card_wt_le {n : ℕ} (C : finset (vector B n)) : Prop := 
  card (filter wt_le C) ≤ 1

instance decidable_pred_card_wt_le {n : ℕ} : 
  decidable_pred (@card_wt_le n) :=
by {intro s, unfold card_wt_le, apply_instance}

def Area1 {n : ℕ} : finset (finset (vector B n)) := 
  filter (card_wt_le) (powerset univ) 

def DelCodes_Area1 {n : ℕ} : finset (finset (vector B n)) := 
  filter is_DelCode Area1

lemma DelCodes_Area1_eq_DelCodes {n : ℕ} :
  @DelCodes_Area1 n = @DelCodes n :=
begin
unfold DelCodes_Area1, unfold DelCodes,
rw subset.antisymm_iff, apply and.intro,
  {intros s sin, rw mem_filter, rw mem_filter at sin,
   apply and.intro,
    {rw mem_powerset, apply subset_univ},
    {apply sin.right}},
  {intros s sin, rw mem_filter, rw mem_filter at sin,
   apply and.intro,
    {unfold Area1, rw mem_filter, apply and.intro,
      {apply sin.left},
      {apply card_filter_wt_le_one s sin.right}},
    {apply sin.right}}
end

def wt_ge {n : ℕ} (x : vector B n) : Prop :=
  wt x ≥ n - 1

instance decidable_pred_wt_ge {n : ℕ} : 
  decidable_pred (@wt_ge n) :=
by {intro s, unfold wt_ge, apply_instance}

lemma card_filter_sub_one_le_wt
  {n : ℕ} (C : finset (vector B n)) (H : is_DelCode C):
  card (filter wt_ge C) ≤ 1 :=
begin
cases decidable.em(filter wt_ge C = ∅) with hem hnem,
  {rw hem, rw card_empty, apply nat.zero_le},
  {cases exists_mem_of_ne_empty hnem with a ain,
   apply le_of_eq, rw card_eq_one, apply exists.intro a,
   rw eq_singleton_iff_unique_mem, apply and.intro,
    {apply ain},
    {intros x xin, 
     rw mem_filter at xin, rw mem_filter at ain,
     apply DelCode'_of_DelCode C H,
      {apply xin.left},
      {apply ain.left},
      {cases sDel_of_le_wt_length_sub x xin.right with i xrepO,
       cases sDel_of_le_wt_length_sub a ain.right with j arepO,
       apply exists.intro i, apply exists.intro j,
       rw xrepO, rw arepO}}}
end

def card_wt_ge {n : ℕ} (C : finset (vector B n)) : Prop := 
  card (filter wt_ge C) ≤ 1

instance decidable_pred_card_wt_ge {n : ℕ} : 
  decidable_pred (@card_wt_ge n) :=
by {intro s, unfold card_wt_ge, apply_instance}

def Area2 {n : ℕ} : finset (finset (vector B n)) := 
  filter (card_wt_ge) (powerset univ) 

def DelCodes_Area2 {n : ℕ} : finset (finset (vector B n)) := 
  filter is_DelCode Area2

lemma DelCodes_Area2_eq_DelCodes {n : ℕ} :
  @DelCodes_Area2 n = @DelCodes n :=
begin
unfold DelCodes_Area2, unfold DelCodes,
rw subset.antisymm_iff, apply and.intro,
  {intros s sin, rw mem_filter, rw mem_filter at sin,
   apply and.intro,
    {rw mem_powerset, apply subset_univ},
    {apply sin.right}},
  {intros s sin, rw mem_filter, rw mem_filter at sin,
   apply and.intro,
    {unfold Area2, rw mem_filter, apply and.intro,
      {apply sin.left},
      {apply card_filter_sub_one_le_wt s sin.right}},
    {apply sin.right}}
end

def Area3 {n : ℕ} : finset (finset (vector B n)) := 
  filter (card_wt_ge) Area1 

def DelCodes_Area3 {n : ℕ} : finset (finset (vector B n)) := 
  filter is_DelCode Area3

lemma DelCodes_Area3_eq_DelCodes {n : ℕ} :
  @DelCodes_Area3 n = @DelCodes n :=
begin
unfold DelCodes_Area3, 
rw ← DelCodes_Area1_eq_DelCodes, unfold DelCodes_Area1, 
rw subset.antisymm_iff, apply and.intro,
  {intros s sin, rw mem_filter, rw mem_filter at sin,
   apply and.intro,
    {unfold Area3 at sin, rw mem_filter at sin, apply sin.left.left},
    {apply sin.right}},
  {intros s sin, rw mem_filter, rw mem_filter at sin,
   apply and.intro,
    {unfold Area3, rw mem_filter, apply and.intro,
      {apply sin.left},
      {apply card_filter_sub_one_le_wt s sin.right}},
    {apply sin.right}}
end

def DelCodes_Area {n : ℕ} : finset (finset (vector B n)) := 
  DelCodes_Area3

lemma DelCodes_Area_eq_DelCodes {n : ℕ} :
  @DelCodes_Area n = @DelCodes n :=
by {unfold DelCodes_Area, apply DelCodes_Area3_eq_DelCodes}

end finset