import data.list.basic data.vector
import .Subsequence 

open list nat prod 

namespace list 

variable {α : Type} 
variables (x : α) (X : list α) (n : ℕ) (N : list ℕ)

local infix `⊴` : 50 := is_subseq

def sDel : list α → ℕ → list α 
| []           _       := []
| (x::[])      _       := []
| (x::x'::X')  0       := x'::X'
| (x::x'::X')  (n + 1) := x :: sDel (x'::X') n 

lemma sDel_nil : 
  sDel ([] : list α) n = [] := 
by trivial

lemma length_sDel_nil : 
  length (sDel ([] : list α) n) = 0 := 
by trivial

lemma sDel_singleton : sDel [x] n = [] := 
by trivial

lemma length_sDel_singleton : 
  length (sDel [x] n) = 0 := 
by rw [sDel_singleton, length]

lemma sDel_zero : sDel (x::X) 0 = X := 
begin
cases X with x' X',
  {trivial},
  {trivial}
end

lemma length_sDel_zero :
  length (sDel (x::X) 0) = length X := 
begin
cases X with x' X',
  {trivial},
  {trivial}
end

lemma sDel_cons
  (H_not_eq_XN : ¬ X = []) :
   sDel (x::X) (n+1) = x :: (sDel X n) := 
begin 
cases X with x'  X', 
  {contradiction},
  {trivial}
end

lemma sDel_cons_of_length (H : 1 ≤ length X) :
   sDel (x::X) (n+1) = x :: (sDel X n) := 
begin 
apply sDel_cons,
assume h, rw h at H, apply nat.not_succ_le_zero 0 H
end

lemma length_sDel :
  length (sDel X n) = length X - 1 := 
begin
revert n, induction X with x X' IHX,
  {intro n, trivial},
  {intro n, cases X' with x' X'',
    {trivial},
    {cases n, 
      {unfold sDel, unfold length, rw nat.add_sub_cancel},
      {unfold sDel, rw length, rw IHX, unfold length, 
       rw nat.add_sub_cancel, rw nat.add_sub_cancel}}}
end

lemma length_sDel_le :
  length (sDel X n) ≤ length X - 1 := 
by rw length_sDel

lemma sDel_subseq : sDel X n ⊴ X := 
begin 
revert X, induction n with n IHn,
  {intro X, cases X with x X',
    {trivial},
    {rw sDel_zero, apply subseq_cons}},
  {intro X, cases X with x X',
    {trivial},
    {cases X' with x' X'',
      {trivial},
      {rw sDel,
       apply cons_subseq_cons x (IHn (x'::X''))}}}
end

lemma sDel_of_ovrlength
  {X : list α} (n : ℕ) (H : length X - 1 ≤ n):
  sDel X n = sDel X (length X) := 
begin
revert n, induction X with x X' IHX,
  {intros n H, repeat {rw sDel_nil}},
  {intros n H, cases X' with x' X'',
    {repeat {rw sDel_singleton}},
    {cases n,
      {unfold length at H, rw nat.add_sub_cancel at H,
       apply false.elim (not_succ_le_zero _ H)},
      {rw length, unfold sDel, rw IHX, 
       apply le_of_succ_le_succ H}}}
end

lemma sDel_of_le_le_length
  {n m : ℕ} {X : list α} 
  (H1 : length X ≤ n) (H2 : length X ≤ m):
  sDel X n = sDel X m := 
begin
revert n m, induction X with x X' IHX,
  {intros n m H1 H2, rw sDel_nil, rw sDel_nil},
  {intros n m H1 H2, cases X' with x' X'',
    {rw sDel_singleton, rw sDel_singleton},
    {cases n,
      {apply false.elim (not_succ_le_zero (length (x'::X'')) H1)},
      {cases m,
        {apply false.elim (not_succ_le_zero (length (x'::X'')) H2)},
        {unfold sDel, rw IHX, 
          {apply le_of_succ_le_succ H1},
          {apply le_of_succ_le_succ H2}}}}}
end

lemma sDel_repeat (a : α) (n i : ℕ) :  
  sDel (repeat a n) i = repeat a (n - 1) :=
begin
revert i, induction n with k IHn,
  {intro i, trivial},
  {intro i, unfold list.repeat, cases k, 
    {trivial},
    {rw nat.succ_sub_one, unfold list.repeat, cases i,
      {unfold sDel},
      {unfold sDel, congr, rw ← repeat_succ, apply IHn}}}
end 

lemma exists_sDel_lt (X : list α) (i : ℕ) :  
  ∃ j : ℕ, j < length X + 1 ∧ sDel X i = sDel X j :=
begin
cases decidable.em(i < length X + 1) with hlt hnlt,
  {apply exists.intro i, apply and.intro,
    {apply hlt},
    {refl}},
  {apply exists.intro (length X), apply and.intro,
    {apply nat.lt_succ_of_le, refl},
    {rw sDel_of_ovrlength, apply le_trans,
      {apply nat.sub_le_self},
      {apply le_trans,
        {apply le_succ},
        {apply le_of_not_gt hnlt}}}}
end 

def sIns : list α → ℕ → α → list α
| []       _       a := [a]
| (x :: X) 0       a := a :: x :: X
| (x :: X) (n + 1) a := x :: sIns X n a 

lemma sIns_nil (a : α) :
 sIns ([] : list α) n a = [a] :=
by trivial

lemma length_sIns_nil (a : α) :
 length (sIns ([] : list α) n a) = 1 :=
by trivial

lemma sIns_zero (a : α) : sIns X 0 a = a :: X := 
begin
cases X with x X, 
  {trivial}, 
  {trivial} 
end

lemma length_sIns_zero (a : α) :
  length (sIns X 0 a) = length X + 1 := 
by {rw sIns_zero, unfold length}

lemma sIns_cons (a : α) :
 sIns (x :: X) (n + 1) a = x :: sIns X n a := 
by trivial

lemma length_sIns (a : α) :
  length (sIns X n a) = length X + 1 := 
begin
revert X, induction n with n IHn,
  {intro X, apply length_sIns_zero},
  {intro X, cases X with x X',
    {trivial},
    {unfold sIns, unfold length, rw IHn}}
end

lemma sIns_neq_nil 
  (X : list α) (n : ℕ) (a : α) :
  ¬ sIns X n a = [] := 
begin
cases X with x X',
  {unfold sIns, contradiction},
  {cases n,
    {unfold sIns, contradiction},
    {unfold sIns, contradiction}}
end

lemma subseq_sIns (a : α) :
  X ⊴ sIns X n a := 
begin
revert X, induction n with n IHn,
  {intro X, rw sIns_zero, apply subseq_cons},
  {intro X, cases X with x X',
    {trivial},
    {rw sIns, apply cons_subseq_cons x (IHn X')}}
end

lemma sIns_of_ovrlength
  (a : α) {n : ℕ} {X : list α} (H : length X ≤ n):
  sIns X n a = X++[a]:= 
begin
revert n, induction X with x X' IHX,
  {intros n H, trivial},
  {intros n H, cases n,
    {apply absurd H, simp},
    {unfold sIns, congr, 
     apply IHX (le_of_succ_le_succ H)}}
end

lemma sIns_of_ovr_ovrlength
  (a : α) {n m : ℕ} {X : list α} 
  (H1 : length X ≤ n) (H2 : length X ≤ m):
  sIns X n a = sIns X m a := 
begin
rw sIns_of_ovrlength a H1,
rw sIns_of_ovrlength a H2
end

lemma sIns_sDel_id
  {X : list α} (n : ℕ) (H : 1 ≤ length X) :
  ∃ (a : α), sIns (sDel X n) n a = X :=
begin
revert n, induction X with x X' IHX,
  {intros n, apply false.elim (not_succ_le_zero 0 H)},
  {intros n, cases X' with x' X'',
    {apply exists.intro x, rw sDel_singleton,
     unfold sIns},
    {cases n,
      {apply exists.intro x, unfold sDel, unfold sIns},
      {cases IHX (le_add_left 1 (length X'')) n with a ha,
       apply exists.intro a,
       unfold sDel, unfold sIns, rw ha}}}
end

lemma sDel_sIns_id
  (n : ℕ) (a : α) :
  sDel (sIns X n a) n = X :=
begin
revert n, induction X with x X' IHX,
  {intros n, trivial},
  {intros n, cases n,
    {trivial},
    {unfold sIns,
     rw sDel_cons,
      {rw IHX},
      {apply sIns_neq_nil}}}
end

lemma sIns_one
    {x a : α} (X : list α) (H : x = a) :
    sIns (x::X) 1 a = sIns (x::X) 0 a :=
begin
cases X with x' X'' ,
    {unfold sIns, rw H},
    {unfold sIns, rw H}
end

lemma sIns_sDel_of_le
  (m : ℕ) (a : α) (H1 : n ≤ length X) (H2 : n ≤ m):
  sIns (sDel (x::X) n) m a  = sDel (sIns (x::X) (m + 1) a) n :=
begin
revert x X m a, induction n with k IHn,
  {intros x X m a H1 H2, rw sDel_zero, 
   unfold sIns, rw sDel_zero},
  {intros x X m a H1 H2, cases X with x' X',
    {apply false.elim (not_succ_le_zero k H1)},
    {unfold sDel, unfold sIns, cases m,
      {apply false.elim (not_succ_le_zero k H2)},
      {unfold sIns, unfold sDel, congr, rw IHn, unfold sIns,
        {apply le_of_succ_le_succ H1},
        {apply le_of_succ_le_succ H2}}}}
end

lemma sIns_sDel_of_le'
  (m : ℕ) (a : α) (H1 : n > length X) (H2 : length X < m):
  sIns (sDel (x::X) n) m a  = sDel (sIns (x::X) m a) (length X) :=
begin
revert x X m a, induction n with k IHn,
  {intros x X m a H1 H2, 
   apply absurd H1, apply not_lt_of_ge, apply nat.zero_le},
  {intros x X m a H1 H2, cases X with x' X',
    {rw sDel_singleton, rw sIns_nil,
     rw sIns_of_ovrlength, 
      {rw length, rw cons_append, rw sDel_zero, rw nil_append},
      {rw length, apply succ_le_of_lt H2}},
    {unfold sDel, cases m,
      {apply false.elim (not_lt_zero (length (x'::X')) H2)},
      {unfold sIns, rw length, rw sDel_cons_of_length,
        {congr, rw IHn, 
          {apply le_of_succ_le_succ H1},
          {apply le_of_succ_le_succ H2}},
        {rw length_sIns, apply succ_le_succ, apply nat.zero_le}}}}
end

lemma sIns_sDel_of_gt
  (m : ℕ) (a : α) (H1 : n ≤ length X) (H2 : n > m):
  sIns (sDel (x::X) n) m a  = sDel (sIns (x::X) m a) (n + 1) :=
begin
revert x X m a, induction n with k IHn,
  {intros x X m a H1 H2, apply false.elim (not_lt_zero m H2)},
  {intros x X m a H1 H2, cases X with x' X',
    {apply false.elim (not_succ_le_zero k H1)},
    {unfold sDel, cases m,
      {unfold sIns, unfold sDel},
      {unfold sIns, cases m,
        {unfold sIns, rw sIns_zero, unfold sDel},
        {unfold sIns, unfold sDel, rw IHn, rw sIns,
        {apply le_of_succ_le_succ H1},
        {apply le_of_succ_le_succ H2}}}}}
end

lemma sIns_sDel_of_gt'
  (m : ℕ) (a : α) (H1 : n > length X) (H2 : length X ≥ m):
  sIns (sDel (x::X) n) m a  = sDel (sIns (x::X) m a) (length X + 2) :=
begin
revert x X m a, induction n with k IHn,
  {intros x X m a H1 H2, 
   apply absurd H1, apply not_lt_of_ge, apply nat.zero_le},
  {intros x X m a H1 H2, cases X with x' X',
    {rw sDel_singleton, rw sIns_nil, 
     have h : m = 0, from nat.eq_zero_of_le_zero H2,
     rw h, rw sIns, rw length, rw sDel, rw sDel_singleton},
    {unfold sDel, cases m,
      {unfold sIns, rw sDel, rw length, rw sDel,
       rw sDel_of_le_le_length, 
        {apply le_of_lt_succ, apply H1},
        {rw length}},
      {unfold sIns, rw length, rw sDel_cons_of_length,
        {congr, rw IHn, 
          {apply le_of_succ_le_succ H1},
          {apply le_of_succ_le_succ H2}},
        {rw length_sIns, apply succ_le_succ, apply nat.zero_le}}}}
end

lemma sIns_sDel_swap (n m : ℕ) (a : α) :
  ∃ n' m' : ℕ, sIns (sDel (x::X) n) m a = sDel (sIns (x::X) m' a) n':=
begin
cases decidable.em(n ≤ length X) with hle hnle,
  {cases decidable.em(n ≤ m) with hle' hnle',
    {apply exists.intro n,
     apply exists.intro (m + 1),
     apply sIns_sDel_of_le,
      {apply hle},
      {apply hle'}},
    {apply exists.intro (n + 1),
     apply exists.intro m,
     apply sIns_sDel_of_gt,
      {apply hle},
      {apply lt_of_not_ge hnle'}}},
  {cases decidable.em(length X < m) with hle' hnle',
    {apply exists.intro (length X),
     apply exists.intro m,
     apply sIns_sDel_of_le',
      {apply lt_of_not_ge hnle},
      {apply hle'}},
    {apply exists.intro (length X + 2),
     apply exists.intro m,
     apply sIns_sDel_of_gt',
      {apply lt_of_not_ge hnle},
      {apply le_of_not_gt hnle'}}},
end

lemma sDel_sIns_swap (n m : ℕ) (a : α) :
  ∃ n' m' : ℕ, ∃ a' : α, sDel (sIns (x::X) n a) m = sIns (sDel (x::X) n') m' a' :=
begin
revert x n m a, induction X with x' X' IHX,
  {intros x n m a, cases n,
    {cases m, 
      {apply exists.intro 0,
       apply exists.intro 1,
       apply exists.intro x,
       trivial},
      {apply exists.intro 0,
       apply exists.intro 1,
       apply exists.intro a,
       trivial}},
    {cases m, 
      {apply exists.intro 0,
       apply exists.intro 1,
       apply exists.intro a,
       trivial},
      {apply exists.intro 0,
       apply exists.intro 0,
       apply exists.intro x,
       trivial}}},
  {intros x n m a, cases n,
    {cases m, 
      {apply exists.intro 0,
       apply exists.intro 0,
       apply exists.intro x,
       trivial},
      {rw sIns, rw sDel,
       apply exists.intro m,
       apply exists.intro 0,
       apply exists.intro a,
       rw sIns_zero}},
    {cases m, 
      {apply exists.intro 0,
       apply exists.intro n,
       apply exists.intro a,
       rw sIns, rw sDel_zero, rw sDel_zero},
      {rw sIns, rw sDel_cons_of_length,
        {cases IHX x' n m a with n' h, 
         cases h with m' h, cases h with a' h,
         apply exists.intro (n' + 1),
         apply exists.intro (m' + 1),
         apply exists.intro a',
         rw sDel, rw sIns, rw h},
        {rw length_sIns, 
         apply succ_le_succ (nat.zero_le (length (x'::X')))}}}}
end

def mDel : list α → list ℕ → list α 
| X []      := X
| X (n :: N):= mDel (sDel X n) N

lemma mDel_nil_left :
  mDel ([] : list α) N = [] := 
begin
induction N with n N' IHN, 
  {trivial},
  {unfold mDel, unfold sDel, apply IHN}
end

lemma mDel_nil_right : mDel X [] = X := 
by trivial

def mDel_of_le {X : list α} {N : list ℕ} (H : length X ≤ length N) : 
  mDel X N = [] :=
begin 
revert X, induction N with n N' IHN,
  {intros X H,
   rw eq_nil_of_length_eq_zero (eq_zero_of_le_zero H),
   rw mDel_nil_right},
  {intros X H, cases X with x X',
    {rw mDel_nil_left},
    {unfold mDel, apply IHN,
     rw length_sDel, unfold length, rw nat.add_sub_cancel,
     apply le_of_succ_le_succ H}}
end

lemma length_mDel :
  length (mDel X N) = length X - length N := 
begin 
revert X, induction N with n N' IHN, 
  {intro X, trivial}, 
  {intro X, unfold mDel, 
   rw [IHN, length_sDel,
       length, nat.sub_sub, nat.add_comm] }
end

lemma mDel_subseq :
  mDel X N ⊴ X := 
begin
revert X, induction N with n N' IHN, 
  {intro X, trivial},
  {intro X, unfold mDel, apply is_subseq.trans 
   (IHN (sDel X n)) (sDel_subseq X n)}
end

lemma mDel_append (X: list α) (N M: list ℕ):
  mDel X (N++M) = mDel (mDel X N) M:=
begin
revert X, induction N with n N' IHN,  
  {intro X, rw nil_append, rw mDel_nil_right},
  {intro X, cases X with x X', 
    {repeat{rw mDel_nil_left}},
    {rw cons_append, unfold mDel, rw IHN}}
end

lemma length_le_of_mDel 
  (X Y : list α) (N : list ℕ) (H : mDel X N = Y) :
  length Y ≤ length X :=
by {apply length_le_of_subseq, rw ← H, apply mDel_subseq}

def mDel_repeat_zero 
  (X : list α) (n : ℕ) (H : length X ≤ n): 
  mDel X (repeat 0 n) = [] :=
by {rw mDel_of_le, rw length_repeat, apply H}

lemma mDel_to_nil (X : list α) :
  mDel X (repeat 0 (length X)) = [] :=
begin
induction X with x X' IH_X,
  {trivial},
  {unfold length, unfold repeat, 
  unfold mDel, rw sDel_zero, apply IH_X}
end

def mDel_append_repeat_zero (X Y: list α) : 
  mDel (X++Y) (repeat 0 (length X)) = Y :=
begin
revert Y, induction X with x X' IHX, 
  {intro Y, trivial},
  {intro Y, rw cons_append, 
   unfold length, unfold repeat,
   unfold mDel, rw sDel_zero, apply IHX}
end

def mDel_repeat_one : 
  mDel (x::X) (repeat 1 (length X)) = [x] :=
begin 
induction X with x' X' IHX,
  {unfold length, unfold repeat, rw mDel_nil_right},
  {unfold length, unfold repeat, 
   unfold mDel, unfold sDel, rw sDel_zero, rw IHX}
end

def succ_list : list ℕ → list ℕ  
| []    := []
| (x::X):= (x + 1) :: succ_list X

def length_succ_list (X : list ℕ):
  length (succ_list X) = length X :=
begin
induction X with x X' IHX,
  {trivial},
  {unfold succ_list, unfold length, rw IHX}
end

lemma mDel_succ_list
  (H_le_NX : length N ≤ length X) :
  mDel (x::X) (succ_list N) = x::mDel X N :=
begin
revert X, induction N with n N' IHN,
  {intros X H_le_NX, trivial},
  {intros X H_le_NX, cases X with x' X',
    {apply false.elim (not_succ_le_zero (length N') H_le_NX)},
    {unfold succ_list, unfold mDel, unfold sDel, rw IHN,
     rw [length_sDel, length, nat.add_sub_cancel] ,
     apply le_of_succ_le_succ H_le_NX}}
end 

lemma mDel_id_of_subseq (X Y : list α):
  X ⊴ Y → ∃ (Z : list ℕ), mDel Y Z = X:=
begin
revert X, induction Y with y Y' IH,
  {intros X H, apply exists.intro [], 
   rw mDel_nil_left, rw nil_of_subseq H},
  {intros X H, cases X with x X',
    {apply exists.intro (repeat 0 (length (y::Y'))),
     rw mDel_repeat_zero, refl},
    {cases H with H_xy H_xXY,
      {cases H_xy with H_xy H_XY, cases (IH X' H_XY) with Z' HZ, 
       cases decidable.em(length Z' ≤ length Y') with h_le h_nle, 
        {apply exists.intro (succ_list Z'),
         rw mDel_succ_list y Y' Z' h_le, rw H_xy, rw HZ},
        {have h : X' = [], 
         rw ← HZ, rw mDel_of_le, apply le_of_lt (lt_of_not_ge h_nle),
         apply exists.intro (repeat 1 (length Y')),
         rw h, rw H_xy, rw mDel_repeat_one}},
      {cases (IH (x::X') H_xXY) with Z' HZ,
       apply exists.intro (0::Z'), 
       unfold mDel, rw sDel_zero, rw HZ}}}
end

lemma subseq_of_mDel_id (X Y : list α):
  (∃ (Z : list ℕ), mDel Y Z = X) → X ⊴ Y :=
by {intro H, cases H with Z HZ, rw ← HZ, apply mDel_subseq}

lemma subseq_iff_mDel_id (X Y : list α):
  X ⊴ Y ↔ ∃ (Z : list ℕ), mDel Y Z = X :=
begin
apply iff.intro,
  {apply mDel_id_of_subseq},
  {apply subseq_of_mDel_id}
end 

lemma mDel_id_length_of_subseq (X Y : list α) (H : X ⊴ Y) : 
  ∃ (Z : list ℕ), length Z ≤ (length Y - length X) ∧ mDel Y Z = X :=
begin
cases X with x X',
  {apply exists.intro (repeat 0 (length Y)),
   apply and.intro,
    {rw length_repeat, refl},
    {rw mDel_repeat_zero, refl}},
  {rw subseq_iff_mDel_id at H, cases H with Z H,
   apply exists.intro Z, apply and.intro,
    {have h : length Y - length Z = length (x::X'),
      {rw ← H, rw length_mDel},
     have h1 : length Z < length Y, 
      {apply nat.lt_of_sub_pos, rw h, apply lt_of_succ_le,
       apply succ_le_succ, apply nat.zero_le},
     rw ← h, rw nat.sub_sub_self, apply le_of_lt h1},
    {rw H}}
end 

lemma subseq_of_mDel_id_length (X Y : list α)
  (H : ∃ (Z : list ℕ), length Z ≤ (length Y - length X) ∧ mDel Y Z = X) :
   X ⊴ Y :=
begin
cases H with Z H,
apply subseq_of_mDel_id,
apply exists.intro Z, apply H.right
end 

lemma subseq_iff_mDel_id_length (X Y : list α):
  X ⊴ Y ↔ 
  ∃ (Z : list ℕ), length Z ≤ (length Y - length X) ∧ mDel Y Z = X :=
begin
apply iff.intro,
  {apply mDel_id_length_of_subseq},
  {apply subseq_of_mDel_id_length}
end 

lemma exists_sDel_of_subseq (X Y : list α) 
  (H1 : is_subseq X Y) (H2 : length Y - length X = 1):
  ∃ i : ℕ, X = sDel Y i :=
begin
rw subseq_iff_mDel_id_length at H1, cases H1 with N H1_N,
rw H2 at H1_N, cases eq_or_lt_of_le H1_N.left with heq hle,
  {rw length_eq_one at heq, cases heq with i H2_i,
   apply exists.intro i, rw H2_i at H1_N, rw ← H1_N.right, refl},
  {have h1 : X ≠ Y,
    {assume h, rw h at H2, rw nat.sub_self at H2, contradiction},
   have h2 : length N = 0,
    {apply nat.eq_zero_of_le_zero, apply nat.le_of_lt_succ hle},
   apply absurd H1_N.right, 
   rw eq_nil_of_length_eq_zero h2, rw mDel_nil_right, apply h1.symm}
end

def mIns : list α → list ℕ → list α → list α 
|   X []       _        :=  X
|   X _        []       :=  X
|   X (y :: Y) (z :: Z) :=  mIns (sIns X y z ) Y Z

lemma mIns_nil_left (X Z: list α) : mIns X [] Z = X := 
begin 
cases Z with z Z', 
    {trivial},
    {trivial}
end

lemma mIns_nil_right (X : list α) (Y : list ℕ) :
  mIns X Y [] = X := 
begin 
cases Y with y Y', 
    {trivial}, 
    {trivial}
end

lemma mIns_take_left (X Y: list α) (N : list ℕ):
  mIns X N Y = mIns X (take (length Y) N) Y :=
begin
revert X N, induction Y with y Y' IHY,  
  {intros X N, rw mIns_nil_right, rw mIns_nil_right},
  {intros X N, cases N with n N', 
    {rw mIns_nil_left, rw take_nil, rw mIns_nil_left},
    {unfold length, unfold take, unfold mIns, apply IHY}}
end

lemma mIns_take_right (X Y: list α) (N : list ℕ):
  mIns X N Y = mIns X N (take (length N) Y) :=
begin
revert X Y, induction N with n N' IHN,  
  {intros X Y, rw mIns_nil_left, rw mIns_nil_left},
  {intros X N, cases N with n N', 
    {rw mIns_nil_right, rw take_nil, rw mIns_nil_right},
    {unfold length, unfold take, unfold mIns, apply IHN}}
end

lemma lentgh_le_mIns (X Y : list α) (N : list ℕ) :
  length X ≤ length (mIns X N Y) := 
begin 
revert X N, induction Y with y Y' IHY,  
  {intros X N, rw mIns_nil_right},
  {intros X N, cases N with n N', 
    {rw mIns_nil_left},
    {unfold mIns, apply le_trans _, 
      {apply IHY}, 
      {rw length_sIns, simp}}}
end

lemma length_mIns_le_right (X Y : list α) (N : list ℕ) :
  length (mIns X N Y) ≤ length X + length Y := 
begin 
revert X N, induction Y with y Y' IHY,  
  {intros X N, rw mIns_nil_right, refl},
  {intros X N, cases N with n N', 
    {rw mIns_nil_left, apply nat.le_add_right},
    {unfold mIns, apply le_trans, 
      {apply IHY}, 
      {rw length_sIns, simp}}}
end

lemma length_mIns_le_left (X Y : list α) (N : list ℕ) :
  length (mIns X N Y) ≤ length X + length N := 
begin 
revert X N, induction Y with y Y' IHY,  
  {intros X N, rw mIns_nil_right, apply nat.le_add_right},
  {intros X N, cases N with n N', 
    {rw mIns_nil_left, refl},
    {unfold mIns, apply le_trans, 
      {apply IHY}, 
      {rw length_sIns, simp}}}
end

lemma length_mIns_of_le (X Y : list α) (N : list ℕ)
  (H : length N ≤ length Y) :
  length (mIns X N Y) = length X + length N := 
begin 
revert X Y, induction N with n N' IHN,  
  {intros X Y Hleq, unfold length, rw mIns_nil_left, simp},
  {intros X Y Hleq, cases Y with y Y', 
    {apply absurd Hleq, apply not_le_of_gt, apply nat.zero_lt_succ},
    {unfold mIns, rw IHN,
      {rw length_sIns, unfold length, simp}, 
      {apply le_of_succ_le_succ Hleq}}}
end

lemma length_mIns_of_ge (X Y : list α) (N : list ℕ)
  (H : length Y ≤ length N) :
  length (mIns X N Y) = length X + length Y := 
begin 
revert X N, induction Y with y Y' IHY,  
  {intros X N Hleq, unfold length, rw mIns_nil_right, simp},
  {intros X N Hleq, cases N with n N', 
    {apply absurd Hleq, apply not_le_of_gt, apply nat.zero_lt_succ},
    {unfold mIns, rw IHY,
      {rw length_sIns, unfold length, simp}, 
      {apply le_of_succ_le_succ Hleq}}}
end

lemma length_mIns (X Y : list α) (N : list ℕ)
  (H : length Y = length N) :
  length (mIns X N Y) = length X + length Y := 
begin 
revert X N, induction Y with y Y' IHY,  
  {intros X N Hleq, unfold length, rw mIns_nil_right, simp},
  {intros X N Hleq, cases N with n N', 
    {contradiction},
    {unfold mIns, rw IHY,
      {rw length_sIns, unfold length, simp}, 
      {apply succ_inj Hleq}}}
end

lemma mIns_append_left (X Y: list α) (N M: list ℕ):
  mIns X (N++M) Y = (mIns (mIns X N Y) M (mDel Y (repeat 0 (length N)))) :=
begin
revert X N M, induction Y with y Y' IHY_iXNMY_eq_iXNY_M,  
  {intros X N M, rw mIns_nil_right, 
   rw mIns_nil_right, rw mDel_nil_left, rw mIns_nil_right},
  {intros X N M, cases N with n N', 
    {rw nil_append, rw mIns_nil_left, unfold length,
     unfold repeat, rw mDel_nil_right},
    {rw cons_append, unfold mIns, rw IHY_iXNMY_eq_iXNY_M,
     unfold length, unfold repeat, 
     unfold mDel, rw sDel_zero}}
end

lemma mIns_append_right (X Y Z: list α) (N : list ℕ):
  mIns X N (Y++Z) = (mIns (mIns X N Y) (mDel N (repeat 0 (length Y))) Z) :=
begin
revert X Y Z, induction N with n N' IHN_iXNYZ_eq_iXNY_Z,  
  {intros X Y Z, rw mIns_nil_left, 
   rw mIns_nil_left, rw mDel_nil_left, rw mIns_nil_left},
  {intros X Y Z, cases Y with y Y', 
    {rw nil_append, rw mIns_nil_right, unfold length,
     unfold repeat, rw mDel_nil_right},
    {rw cons_append, unfold mIns, rw IHN_iXNYZ_eq_iXNY_Z,
     unfold length, unfold repeat, 
     unfold mDel, rw sDel_zero}}
end

lemma mIns_append_left_of_le
  (X Y: list α) (N M: list ℕ) (H : length Y ≤ length N):
  mIns X (N++M) Y = mIns X N Y  :=
begin
rw mIns_append_left, rw mDel_repeat_zero, 
  {rw mIns_nil_right},
  {apply H}
end

lemma mIns_append_right_of_le
  (X Y Z: list α) (N: list ℕ) (H : length N ≤ length Y):
  mIns X N (Y++Z) = mIns X N Y :=
begin
rw mIns_append_right, rw mDel_repeat_zero, 
  {rw mIns_nil_left},
  {apply H}
end

lemma mIns_append_append 
  (X Y Z: list α) (N M: list ℕ) (H : length N = length Y):
  mIns X (N++M) (Y++Z) = mIns (mIns X N Y) M Z :=
begin
rw mIns_append_left, rw mIns_append_right,
rw mDel_repeat_zero,
  {rw mIns_nil_left, rw H, rw mDel_append_repeat_zero},
  {rw H}
end

lemma mIns_repeat_zero (X Y : list α) (n : ℕ):
  mIns X (repeat 0 n) Y = reverse (take n Y) ++ X:=
begin
revert X Y, induction n with k IHk,
  {intros X Y, unfold repeat, rw mIns_nil_left,
   rw take_zero, rw reverse_nil, rw nil_append},
  {intros X Y, cases Y with y Y',
    {rw mIns_nil_right, rw take_nil, rw reverse_nil, rw nil_append},
    {unfold repeat, unfold mIns, rw sIns_zero, rw IHk, simp}}
end

lemma mIns_from_nil (X : list α) :
  mIns [] (repeat 0 (length X)) (reverse X) = X :=
begin
rw mIns_repeat_zero, rw take_all_of_le,
  {rw reverse_reverse, rw append_nil},
  {rw length_reverse}
end

lemma mIns_mDel_id_le
  {X : list α} (N : list ℕ) (H : length N ≤ length X):
  ∃ (Y : list α), length Y = length N ∧ mIns (mDel X N) (reverse N) Y = X :=
begin
revert X, induction N with n N' IHN,
  {intros X H, apply exists.intro [], 
   rw mDel_nil_right, rw mIns_nil_right, simp},
  {intros X H, cases X with x X',
    {apply absurd H, simp},
    {unfold mDel, 
     cases @IHN (sDel (x::X') n) _ with Y IHN_Y,
      {have h : ∃ y : α, sIns (sDel (x::X') n) n y = (x::X'),
        {apply sIns_sDel_id, 
         apply succ_le_succ (nat.zero_le (length X'))}, 
         cases h with y IH_y, apply exists.intro (Y++[y]),
         apply and.intro,
          {rw length_append, rw IHN_Y.left, simp},
          {rw reverse_cons, rw mIns_append_right,
           rw ← length_reverse N' at IHN_Y, rw IHN_Y.left, 
           rw mDel_append_repeat_zero,
           rw mIns_append_left_of_le, 
            {rw IHN_Y.right, unfold mIns, rw IH_y},
            {rw IHN_Y.left}}},
      {rw length_sDel, unfold length,
       rw nat.add_sub_cancel, apply le_of_succ_le_succ H}}}
end

lemma mIns_mDel_id_ge
  {X : list α} (N : list ℕ) (H : length X ≤ length N):
  mIns (mDel X N) (repeat 0 (length X)) (reverse X) = X :=
by {rw mDel_of_le H, rw mIns_from_nil}

lemma mIns_mDel_id
  {X : list α} (N : list ℕ) :
  ∃ (M : list ℕ) (Y : list α), mIns (mDel X N) M Y = X :=
begin
cases decidable.em(length N ≤ length X) with h_le h_lt,
  {have h : ∃ (Y : list α), 
    length Y = length N ∧ mIns (mDel X N) (reverse N) Y = X,
    from mIns_mDel_id_le N h_le,
   cases h with Y h_Y,
   apply exists.intro (reverse N), apply exists.intro Y, 
   apply h_Y.right},
  {apply exists.intro (repeat 0 (length X)),
   apply exists.intro (reverse X),
   rw mIns_mDel_id_ge, apply le_of_lt (lt_of_not_ge h_lt)}
end

lemma mDel_mIns_id_le
  (X Y: list α) (N : list ℕ) (H : length N ≤ length Y):
  mDel (mIns X N Y) (reverse N) = X :=
begin
revert X Y, induction N with n N' IHN,
  {intros X Y H, rw mIns_nil_left, 
   rw reverse_nil, rw mDel_nil_right},
  {intros X Y H, cases Y with y Y',
    {apply absurd H, simp},
    {unfold mIns, rw reverse_cons, rw mDel_append, rw IHN _,
      {unfold mDel, rw sDel_sIns_id},
      {apply le_of_succ_le_succ H}}}
end

lemma mDel_mIns_id_ge
  (X Y: list α) (N : list ℕ) (H : length Y ≤ length N):
  mDel (mIns X N Y) (reverse (take (length Y) N)) = X :=
begin
revert X Y, induction N with n N' IHN,
  {intros X Y H, rw mIns_nil_left, 
   rw take_nil, rw reverse_nil, rw mDel_nil_right},
  {intros X Y H, cases Y with y Y',
    {rw mIns_nil_right, unfold length, 
     rw take_zero, rw reverse_nil, rw mDel_nil_right},
    {unfold mIns, unfold length, rw take_cons,
     rw reverse_cons, rw mDel_append, rw IHN,
      {unfold mDel, rw sDel_sIns_id},
      {apply le_of_succ_le_succ H}}}
end

lemma mDel_mIns_id
  (X Y: list α) (N : list ℕ):
  ∃ M : list ℕ, mDel (mIns X N Y) M = X :=
begin
cases decidable.em(length N ≤ length Y) with h_le h_lt,
  {apply exists.intro (reverse N),
   rw mDel_mIns_id_le X Y N h_le},
  {apply exists.intro (reverse (take (length Y) N)),
   rw mDel_mIns_id_ge X Y N (le_of_lt (lt_of_not_ge h_lt))}
end

lemma subseq_mIns (X Z : list α) (Y : list ℕ):
  X ⊴ mIns X Y Z := 
begin
revert X Z, induction Y with y Y' IHY_XiXYZ, 
  {intros X Z, rw mIns_nil_left},
  {intros X Z, cases Z with z Z',
    {rw mIns_nil_right}, 
    {unfold mIns,
     apply is_subseq.trans _ (IHY_XiXYZ (sIns X y z) Z'),
     apply subseq_sIns}}
end

lemma mIns_id_of_subseq (X Y : list α) :
  X ⊴ Y → ∃ (N : list ℕ) (Z : list α), mIns X N Z = Y :=
begin
intro H, rw subseq_iff_mDel_id at H,
cases H with M H_M, rw ← H_M, 
apply mIns_mDel_id
end

lemma subseq_of_mIns_id (X Y : list α):
  (∃ (N : list ℕ) (Z : list α), mIns X N Z = Y) → X ⊴ Y :=
begin
intro H, cases H with N HN, cases HN with Z HNZ,
rw ← HNZ, apply subseq_mIns
end

lemma subseq_iff_mIns_id (X Y : list α):
  X ⊴ Y ↔ ∃ (N : list ℕ) (Z : list α), mIns X N Z = Y :=
begin
apply iff.intro,
  {apply mIns_id_of_subseq},
  {apply subseq_of_mIns_id}
end 

lemma mIns_id_length_of_subseq (X Y : list α) (H : X ⊴ Y) :
   ∃ (N : list ℕ) (Z : list α), 
   length N ≤ (length Y - length X) ∧ mIns X N Z = Y :=
begin
rw subseq_iff_mIns_id at H, 
cases H with N H,  cases H with Z H,
apply exists.intro (take (length Z) N), apply exists.intro Z,
apply and.intro,
  {rw length_take, 
   cases decidable.em(length Z ≤ length N) with hle hnle,
    {unfold min, rw if_pos hle, rw ← H,
     rw length_mIns_of_ge X Z N hle, 
     rw nat.add_sub_cancel_left},
    {unfold min, rw if_neg hnle, rw ← H,
     rw length_mIns_of_le X Z N,
      {rw nat.add_sub_cancel_left},
      {apply le_of_lt, apply lt_of_not_ge hnle}}},
    {rw ← mIns_take_left, rw H}
end 

lemma subseq_of_mIns_id_length (X Y : list α) 
  (H : ∃ (N : list ℕ) (Z : list α), 
   length N ≤ (length Y - length X) ∧ mIns X N Z = Y) :
   X ⊴ Y :=
begin
cases H with N H, cases H with Z H,
apply subseq_of_mIns_id,
apply exists.intro N, apply exists.intro Z, 
apply H.right
end 

lemma subseq_iff_mIns_id_length (X Y : list α):
  X ⊴ Y ↔ 
  ∃ (N : list ℕ) (Z : list α), 
  length N ≤ (length Y - length X) ∧ mIns X N Z = Y :=
begin
apply iff.intro,
  {apply mIns_id_length_of_subseq},
  {apply subseq_of_mIns_id_length}
end 

lemma sIns_subseq_of_subseq 
  (X Y : list α)(n : ℕ) (H1 : X ⊴ Y) (H2 : length X < length Y) :
  ∃ (i : ℕ), ∃ (a : α), (sIns X i a) ⊴ Y := 
begin 
have h : ∃ (N : list ℕ), ∃ (Z : list α), mIns X N Z = Y,
apply mIns_id_of_subseq X Y H1,
cases h with N h_N, cases h_N with Z HN_Z,
cases N with n N',
  {rw mIns_nil_left at HN_Z, rw HN_Z at H2, 
   apply absurd H2, apply not_lt_of_ge, rw ge},
  {cases Z with z Z',
    {rw mIns_nil_right at HN_Z, rw HN_Z at H2, 
     apply absurd H2, apply not_lt_of_ge, rw ge},
    {apply exists.intro n, apply exists.intro z,
     apply subseq_of_mIns_id, 
     apply exists.intro N', apply exists.intro Z',
     unfold mIns at HN_Z, rw HN_Z}}
end

lemma mIns_of_mDel 
  (X Y : list α) (H : ∃ N : list ℕ, mDel X N = Y) :
  ∃ N : list ℕ, ∃ Z : list α, mIns Y N Z = X  := 
begin 
apply mIns_id_of_subseq,
apply subseq_of_mDel_id Y X H
end

lemma mDel_of_mIns (X Y : list α) 
  (H : ∃ N : list ℕ, ∃ Z : list α, mIns Y N Z = X) :
  ∃ N : list ℕ, mDel X N = Y  := 
begin 
apply mDel_id_of_subseq,
apply subseq_of_mIns_id Y X H
end

lemma mIns_iff_mDel (X Y : list α) :
  (∃ N : list ℕ, ∃ Z : list α, mIns Y N Z = X)
  ↔ (∃ N : list ℕ, mDel X N = Y) := 
begin 
apply iff.intro,
  {apply mDel_of_mIns},
  {apply mIns_of_mDel}
end

lemma sIns_sIns (n m : ℕ)(a b : α)(X : list α):
  sIns (sIns X n a) m b = mIns X (n::[m]) (a::[b]) :=
by trivial

lemma mIns_sIns (n : ℕ)(a : α)(X Y: list α)(N : list ℕ):
  mIns (sIns X n a) N Y = mIns X (n::N) (a::Y) :=
by trivial

lemma sDel_sDel (n m : ℕ) (X : list α):
  sDel (sDel X n) m = mDel X (n::[m]) :=
by trivial

lemma mDel_sDel (n : ℕ) (X : list α) (N : list ℕ):
  mDel (sDel X n) N = mDel X (n::N) :=
by trivial

end  list 

namespace vector 

variable {α : Type} 

lemma list.length_sDel {n : ℕ} (X : vector α n) (i : ℕ) :
  list.length (list.sDel (to_list X) i) = n - 1 := 
by {rw list.length_sDel, rw to_list_length}

def sDel {n : ℕ} (X : vector α n) (i : ℕ) : vector α (n - 1) := 
  subtype.mk (list.sDel (to_list X) i) (list.length_sDel X i)

lemma sDel_nil {i : ℕ} : 
  sDel (@vector.nil α) i = vector.nil := 
by trivial

lemma length_sDel_nil {i : ℕ}: 
  length (sDel (@vector.nil α) i) = 0 := 
by trivial

lemma sDel_singleton (X : vector α 1) (i : ℕ) : 
  sDel X i = vector.nil := 
by rw vector.eq_nil (sDel X i)

lemma length_sDel_singleton (X : vector α 1) (i : ℕ) : 
  length (sDel X i) = 0 := 
by rw sDel_singleton

lemma sDel_zero {n : ℕ} (X : vector α n) (x : α) :
  sDel (x::X) 0 = X := 
begin
unfold sDel, rw subtype.ext,
repeat {rw ← to_list}, rw to_list_mk, 
rw to_list_cons, rw list.sDel_zero
end

lemma length_sDel_zero {n : ℕ} (X : vector α n) (x : α) :
  length (sDel (x::X) 0) = n := 
by {rw sDel_zero, refl}

lemma sDel_cons {n : ℕ} (X : vector α (n + 1)) (x : α) (i : ℕ) :
   sDel (x::X) (i + 1) = x :: (sDel X i) := 
begin 
unfold sDel, rw subtype.ext,
rw ← to_list, rw to_list_mk, rw to_list_cons, 
rw ← to_list, rw to_list_cons, rw list.sDel_cons,
  {congr},
  {assume h, 
   have h1 : list.length (to_list X) = 0,
    {rw h, refl},
   apply absurd h1, apply ne_of_gt,
   rw to_list_length, apply lt_succ_of_le, apply nat.zero_le}
end

lemma length_sDel {n : ℕ} (X : vector α n) (i : ℕ):
  length (sDel X i) = n - 1 := 
by unfold sDel

lemma length_sDel_le {n : ℕ} (X : vector α n) (i : ℕ):
  length (sDel X i) ≤ n - 1 := 
by rw length_sDel

lemma sDel_repeat (a : α) (n i : ℕ) :  
  sDel (repeat a n) i = repeat a (n - 1) :=
by {rw subtype.ext, unfold sDel, apply list.sDel_repeat}

lemma exists_sDel_of_subseq 
  {n : ℕ} (X : vector α n) (Y : vector α (n - 1))
  (H : is_subseq (to_list Y) (to_list X)) :
  ∃ i : ℕ, Y = sDel X i :=
begin
cases n,
  {apply exists.intro 0, 
     rw vector.eq_nil X, rw vector.sDel_nil, rw vector.eq_nil Y},
  {cases list.exists_sDel_of_subseq (to_list Y) (to_list X) H _ with i h_i, 
    {apply exists.intro i, unfold sDel, rw subtype.ext, apply h_i},
    {repeat {rw to_list_length}, rw succ_sub_one, rw nat.succ_sub,
      {rw nat.sub_self},
      {apply le_refl}}}
end

lemma subseq_of_exists_sDel
  {n : ℕ} (X : vector α n) (Y : vector α (n - 1))
  (H : ∃ i : ℕ, Y = sDel X i) :
  is_subseq (to_list Y) (to_list X):=
by {cases H with i H_i, rw H_i, apply list.sDel_subseq}

lemma subseq_iff_exists_sDel
  {n : ℕ} (X : vector α n) (Y : vector α (n - 1)) :
  is_subseq (to_list Y) (to_list X) ↔ ∃ i : ℕ, Y = sDel X i :=
begin
apply iff.intro,
  {apply exists_sDel_of_subseq},
  {apply subseq_of_exists_sDel}
end

lemma exists_sDel_lt {n : ℕ} (X : vector α n) (i : ℕ) :  
  ∃ j : ℕ, j < n + 1 ∧ sDel X i = sDel X j :=
begin
cases list.exists_sDel_lt (to_list X) i with j h_j,
  {apply exists.intro j, apply and.intro,
    {rw to_list_length at h_j, apply h_j.left},
    {unfold sDel, rw subtype.ext, apply h_j.right}}
end 

lemma exists_sDel_range {n : ℕ} (X : vector α n) (i : ℕ) :  
  ∃ j ∈ range (n + 1), sDel X i = sDel X j :=
begin
cases vector.exists_sDel_lt X i with j h_j,
  {apply exists.intro j, apply exists.intro,
    {rw mem_range, apply h_j.left},
    {apply h_j.right}}
end 

end vector 
