import .InsDel
import .B
import .NumOsNumIs
import .Moment
import .SubMod
import data.vector

open list nat vector

def LevCode (n a m : ℕ) : set (vector B n) :=
  {X| (ρ (to_list X)) % m = a % m ∧ (n + 1) ≤ m}

namespace LevenshteinCode

namespace list 

def def_in_LevCode {n a m : ℕ} {X : vector B n} :
    X ∈ LevCode n a m 
    ↔ (ρ (to_list X)) % m = a % m ∧ n + 1 ≤ m :=
by trivial

def in_LevCode_iff_mod
    {n a₁ m : ℕ} {X : vector B n} (a₂ : ℕ) 
    (H_LCa₁mn : X ∈ LevCode n a₁ m) :
    X ∈ LevCode n a₂ m ↔ a₁ % m = a₂ % m :=
begin
apply iff.intro, 
    {intro H_LCa₂mn, rw ← H_LCa₁mn.left, rw ← H_LCa₂mn.left},
    {intro H, unfold LevCode, rw ← H, apply H_LCa₁mn}
end


lemma sIns_fig_of_pos
    {n a m i : ℕ} {X : vector B n} {Y : vector B (n - 1)} 
    (H1 : X ∈ LevCode n a m) 
    (H2 : to_list Y = sDel (to_list X) i)
    (H3 : 1 ≤ n)
    (H4 : sub_mod a (to_list Y) m ≤ num_Is (to_list Y)) :
    to_list X = sIns (to_list Y) i O :=
begin
rw H2, rw H2 at H4, 
apply sIns_fig_of_pos H1.left _ _ H4,
    {rw to_list_length, apply H1.right},
    {rw to_list_length, apply H3}
end

lemma sub_mod_sDel_of_pos
    {n a m i : ℕ} {X : vector B n} {Y : vector B (n - 1)} 
    (H1 : X ∈ LevCode n a m) 
    (H2 : to_list Y = sDel (to_list X) i) 
    (H3 : 1 ≤ n) 
    (H4 : sub_mod a (to_list Y) m ≤ num_Is (to_list Y)) :
    sub_mod a (to_list Y) m = num_RIs (to_list X) i :=
begin
rw H2, rw sub_mod_sDel_of_pos,
    {rw H1.left},
    {rw to_list_length, apply H1.right},
    {rw to_list_length, apply H3},
    {rw ← H2, apply H4}
end

lemma sIns_fig_of_neg
    {n a m i : ℕ} {X : vector B n} {Y : vector B (n - 1)} 
    (H1 : X ∈ LevCode n a m) 
    (H2 : (to_list Y) = sDel (to_list X) i)
    (H3 : 1 ≤ n)
    (H4 : ¬ sub_mod a (to_list Y) m ≤ num_Is (to_list Y)) :
    to_list X = sIns (to_list Y) i I :=
begin
rw H2, rw H2 at H4,
apply sIns_fig_of_neg H1.left _ _ H4,
    {rw to_list_length, apply H1.right},
    {rw to_list_length, apply H3}
end

lemma sub_mod_sDel_of_neg
    {n a m i : ℕ} {X : vector B n} {Y : vector B (n - 1)} 
    (H1 : X ∈ LevCode n a m) 
    (H2 : (to_list Y) = sDel (to_list X) i)
    (H3 : 1 ≤ n)
    (H4 : ¬ sub_mod a (to_list Y) m ≤ num_Is (to_list Y)) : 
    sub_mod a (to_list Y) m = num_LOs (to_list X) i + num_Is (to_list Y) + 1 :=
begin
rw H2, rw sub_mod_sDel_of_neg,
    {rw H1.left},
    {rw to_list_length, apply H1.right},
    {rw to_list_length, apply H3},
    {rw ← H2, apply H4}
end

-- decoding algorithm

def decoding_alg (n a m : ℕ) (Y : list B) :=
if list.length Y = n then Y 
    else if sub_mod a Y m  ≤ num_Is Y
        then sIns Y (max_num_RIs Y (sub_mod a Y m)) O
        else sIns Y (min_num_LOs Y ((sub_mod a Y m) - num_Is Y - 1)) I

example : decoding_alg 5 5 8 [I, I, O, O] = [O, I, I, O, O] := rfl
example : decoding_alg 5 5 8 [O, I, O, O] = [O, I, I, O, O] := rfl
example : decoding_alg 5 5 8 [O, I, I, O] = [O, I, I, O, O] := rfl

lemma length_decoding_alg (n a m : ℕ) (Y : vector B (n - 1)) :
    list.length (decoding_alg n a m (to_list Y)) = n :=
begin 
unfold decoding_alg, 
cases decidable.em (n ≤ 0) with hle_len hnle_len,
    {have h : list.length (to_list Y) = n,
        {rw to_list_length, rw eq_zero_of_le_zero hle_len}, 
     rw if_pos h, rw h},
    {have h : list.length (to_list Y) ≠ n,
        {apply ne_of_lt, rw to_list_length, 
         apply nat.sub_lt_self,
            {apply lt_of_not_ge hnle_len},
            {apply zero_lt_one}},
     rw if_neg h, 
     cases decidable.em(sub_mod a (to_list Y) m ≤ num_Is (to_list Y)) with hle_wt hnle_wt,
        {rw if_pos hle_wt, rw length_sIns, rw to_list_length, 
         rw nat.sub_add_cancel, apply nat.succ_le_of_lt, apply lt_of_not_ge hnle_len},
        {rw if_neg hnle_wt, rw length_sIns, rw to_list_length, 
         rw nat.sub_add_cancel, apply nat.succ_le_of_lt, apply lt_of_not_ge hnle_len}}
end

end list

namespace vector 

def decoding_alg (n a m : ℕ) (Y : vector B (n - 1)) : vector B n :=
    ⟨list.decoding_alg n a m (to_list Y), list.length_decoding_alg n a m Y⟩ 

def decoding_alg_to_list (n a m : ℕ) (Y : vector B (n - 1)) : 
    to_list (decoding_alg n a m Y) = list.decoding_alg n a m (to_list Y) :=
by trivial

lemma sDelErr_correctable_of_nil
    {a m i : ℕ} {X : vector B 0} {Y : vector B (0 - 1)}
    (H1 : X ∈ LevCode 0 a m) (H2 : to_list Y = sDel (to_list X) i) :
    decoding_alg 0 a m Y = X := 
begin
rw subtype.ext, repeat {rw ← to_list}, 
rw decoding_alg_to_list,
unfold list.decoding_alg, rw if_pos, 
    {rw vector.eq_nil X, rw vector.eq_nil Y},
    {rw to_list_length}
end

lemma sDelErr_correctable_of_pos
    {n a m i : ℕ} {X : vector B n} {Y : vector B (n - 1)}
    (H1 : X ∈ LevCode n a m) (H2 : to_list Y = sDel (to_list X) i)
    (H3 : 1 ≤ n) (H4 : sub_mod a (to_list Y) m ≤ num_Is (to_list Y)) :
    decoding_alg n a m Y = X := 
begin
rw subtype.ext, repeat {rw ← to_list}, 
rw decoding_alg_to_list,
unfold list.decoding_alg, rw if_neg, 
    {rw if_pos H4,
     rw list.sIns_fig_of_pos H1 H2 H3 H4, 
     apply sIns_zero_of_num_RIs,
     rw num_RIs_sIns_max_num_RIs_id O H4,
     rw list.sub_mod_sDel_of_pos,
        {rw ← list.sIns_fig_of_pos H1 H2 H3 H4},
        {apply H1},
        {apply H2},
        {apply H3},
        {apply H4}},
    {apply ne_of_lt, rw H2, rw list.length_sDel, 
     rw to_list_length, apply nat.sub_lt, 
        {apply lt_of_succ_le H3},
        {apply zero_lt_one}}
end

lemma sDelErr_correctable_of_neg
    {n a m i : ℕ} {X : vector B n} {Y : vector B (n - 1)}
    (H1 : X ∈ LevCode n a m) (H2 : to_list Y = sDel (to_list X) i)
    (H3 : 1 ≤ n) (H4 : ¬ sub_mod a (to_list Y) m ≤ num_Is (to_list Y))  :
    decoding_alg n a m Y = X := 
begin
rw subtype.ext, repeat {rw ← to_list}, 
rw decoding_alg_to_list,
unfold list.decoding_alg, rw if_neg, 
    {rw if_neg H4,
     rw list.sIns_fig_of_neg H1 H2 H3 H4, 
     apply sIns_one_of_num_LOs,
     rw num_LOs_sIns_min_num_LOs_id,
        {rw nat.sub_sub, rw list.sub_mod_sDel_of_neg H1 H2 H3 H4,
         rw ← list.sIns_fig_of_neg H1 H2 H3 H4, 
         rw add_assoc, rw nat.add_sub_cancel},
        {rw nat.sub_sub, rw list.sub_mod_sDel_of_neg H1 H2 H3 H4,
         rw add_assoc, rw nat.add_sub_cancel,
         rw list.sIns_fig_of_neg H1 H2 H3 H4,
         apply le_trans,
            {apply num_LOs_le_num_Os},
            {rw num_Os_sIns_one}}},
    {apply ne_of_lt, rw H2, rw list.length_sDel, 
    rw to_list_length, apply nat.sub_lt, 
        {apply lt_of_succ_le H3},
        {apply zero_lt_one}}
end

theorem sDelErr_correctable
  {n a m : ℕ} {X ∈ LevCode n a m} 
  {i : ℕ} {Y : vector B (n - 1)} (h : Y = sDel X i) :
  decoding_alg n a m Y = X :=
begin
rw subtype.ext at h, cases n,
  {rw sDelErr_correctable_of_nil H h},
  {have hle : 1 ≤ succ n, from succ_le_succ (nat.zero_le n),
   cases decidable.em (sub_mod a (to_list Y) m ≤ num_Is (to_list Y)) with hp hn,
    {rw sDelErr_correctable_of_pos H h hle hp},
    {rw sDelErr_correctable_of_neg H h hle hn}}
end

end vector

end LevenshteinCode

