import .InsDel
import .NumOsNumIs

open list nat

namespace list 

def moment: list B → ℕ
| []     := 0 
| (x::X) := moment X + num_Is (x::X)

#eval moment [O,I,I]
#eval moment [O,I,O,O,I]
#eval moment [O,I,O,O,I,I,O]

notation `ρ` X := moment X

#eval ρ [O,I,O,O,I]
#eval ρ [O,I,O,O,I,I,O]

-- original definition of moment

def moment_original_sub : list B → ℕ → ℕ 
| []     _ := 0 
| (O::X) n := moment_original_sub X (n + 1)
| (I::X) n := moment_original_sub X (n + 1) + n

lemma moment_original_sub_succ
    (X : list B) (n : ℕ) :
    moment_original_sub X (n + 1) = moment_original_sub X n + num_Is X :=
begin
revert n, induction X with x X' IHX,
    {intro n, trivial},
    {intro n, cases x,
        {unfold moment_original_sub, unfold num_Is, rw IHX},
        {unfold moment_original_sub, unfold num_Is, rw IHX,
         repeat {rw add_assoc}, rw add_left_cancel_iff, rw add_left_comm}}
end

def moment_original (X : list B) :=  moment_original_sub X 1

lemma moment_original_nil :
    moment_original [] = 0 :=
by trivial

lemma moment_original_singleton
    (x : B):
    moment_original [x] = num_Is [x] :=
begin
cases x,
    {trivial},
    {trivial}
end

lemma moment_original_cons
    (x : B) (X : list B) :
    moment_original (x::X) = moment_original X + num_Is (x::X) :=
begin
revert x, cases X with x' X',
    {intro x, rw moment_original_singleton,
     rw moment_original_nil, rw nat.zero_add},
    {intro x, unfold moment_original, cases x,
        {unfold moment_original_sub, unfold num_Is,
         apply moment_original_sub_succ},
        {unfold moment_original_sub, unfold num_Is, rw ← add_assoc,
         rw moment_original_sub_succ}}
end

lemma moment_iff_moment_original 
    (X : list B) :
    moment X = moment_original X :=
begin
induction X with x X' IHX,
    {trivial},
    {unfold moment, rw moment_original_cons, rw IHX}
end


-- moment

lemma moment_nil:
    (ρ []) = 0 := 
by trivial

lemma moment_zero:
    (ρ [O]) = 0 := 
by trivial

lemma moment_one :
    (ρ [I]) = 1 := 
by trivial

lemma moment_cons (x : B) (X : list B):
    (ρ x::X) = (ρ X) + num_Is (x::X) := 
by trivial

lemma moment_le_cons
    (x : B) (X : list B) :
    (ρ X) ≤  (ρ x::X) := 
by {unfold moment, simp}

lemma moment_sDel_le
    (X : list B) (i : ℕ) :
    (ρ sDel X i) ≤ (ρ X) := 
begin
revert i, induction X with x X' IHX,
    {intro i, trivial},
    {intro i, cases X' with x' X'',
        {apply nat.zero_le},
        {cases i,
            {unfold sDel, apply moment_le_cons},
            {unfold sDel, rw moment, rw moment, 
             apply add_le_add (IHX i), cases x,
                {unfold num_Is, apply num_Is_sDel_le},
                {unfold num_Is, apply succ_le_succ, 
                 apply num_Is_sDel_le}}}}
end

lemma moment_le_sIns 
    (X : list B) (i : ℕ) (a : B) :
    (ρ X) ≤ (ρ sIns X i a) := 
begin
revert i, induction X with x X' IHX,
    {intro i, unfold moment, apply nat.zero_le},
    {intro i, cases i,
        {unfold sIns, apply moment_le_cons},
        {unfold sIns, unfold moment, 
         apply add_le_add (IHX i), cases x,
            {unfold num_Is, apply num_Is_le_sIns},
            {unfold num_Is, apply succ_le_succ, 
             apply num_Is_le_sIns}}}
end

lemma moment_sIns_zero
    (X : list B) (i : ℕ) :
    (ρ sIns X i O) = (ρ X) + num_RIs (sIns X i O) i := 
begin
revert i, induction X with x X' IHX,
    {intro i, unfold sIns, unfold moment, unfold num_Is,
     rw num_RIs_singleton},
    {intro i, cases i,
        {trivial},
        {unfold sIns, unfold moment, unfold num_RIs, 
         rw num_Is_cons, rw num_Is_cons x X', rw IHX,
         rw num_Is_sIns_zero, simp}}
end

-- We should make more more general lemma for calc

lemma moment_sIns_one
    {X : list B} {i : ℕ} :
    (ρ sIns X i I) = (ρ X) + num_LOs (sIns X i I) i + num_Is X + 1 := 
begin
revert i, induction X with x X' IHX,
    {intro i, unfold sIns, rw moment_one,
     unfold moment, cases i,
        {unfold num_LOs, unfold num_Is},
        {unfold num_LOs, unfold num_Is}},
    {intro i, cases i,
        {unfold sIns, rw moment, unfold num_Is, 
         unfold num_LOs, simp},
        {unfold sIns, unfold moment, rw IHX, cases x,
            {unfold num_Is, rw num_Is_sIns_one,
             rw num_LOs, simp},
            {unfold num_Is, rw num_Is_sIns_one,
             rw num_LOs, rw ← add_assoc, simp}}}
end

lemma moment_sub_sDel_total
    {X : list B} (i : ℕ) (H : 1 ≤ length X):
    ((ρ X) - (ρ sDel X i) = num_RIs X i ∧ X = sIns (sDel X i) i O)
    ∨ ((ρ X) - (ρ sDel X i) = num_LOs X i + num_Is (sDel X i) + 1 ∧ X = sIns (sDel X i) i I) :=
begin
have h : ∃ (a : B), sIns (sDel X i) i a = X,
apply sIns_sDel_id i H,
cases h with a ha, cases a,
    {rw ← ha, rw sDel_sIns_id, left,
     apply and.intro,
        {rw moment_sIns_zero, rw nat.add_sub_cancel_left},
        {refl}},
    {rw ← ha, rw sDel_sIns_id , right,
     apply and.intro,
        {rw moment_sIns_one, repeat{rw add_assoc}, rw add_comm, 
         repeat{rw ← add_assoc}, rw nat.add_sub_cancel},
        {refl}}
end

lemma moment_sub_sDel_le
    (X : list B) (i : ℕ) (H : 1 ≤ length X) :
    (ρ X) - (ρ sDel X i) ≤ length X :=
begin
revert i, cases X with x X',
    {intro i, apply nat.zero_le},
    {intro i, cases moment_sub_sDel_total i H,
        {cases h with hl hr, rw hl, 
         apply le_trans (num_RIs_le_length (x::X') i), apply nat.sub_le},
        {cases h with hl hr, rw hl, rw hr,
         cases decidable.em(i ≤ length (sDel (x::X') i)) with h_le h_nle,
            {rw num_LOs_sIns I h_le, rw sDel_sIns_id,
             rw length_sIns, apply succ_le_succ, 
             rw ← num_Os_add_num_Is,
             apply add_le_add,
                {apply num_LOs_le_num_Os},
                {refl}},
            {rw num_LOs_of_ovrlen, 
                {rw num_Os_sIns_one, rw sDel_sIns_id, rw length_sIns,
                 apply succ_le_succ, rw ← num_Os_add_num_Is},
                {rw length_sIns, apply succ_le_of_lt,
                 apply lt_of_not_ge h_nle}}}}
end 

lemma sIns_fig_of_pos_of_moment
    {X : list B} {i : ℕ}
    (H1 : 1 ≤ length X)
    (H2 : (ρ X) - (ρ sDel X i) ≤ num_Is (sDel X i)) :
    X = sIns (sDel X i) i O :=
begin
cases moment_sub_sDel_total i H1,
    {apply h.right},
    {apply absurd h.left, apply ne_of_lt,
     apply lt_of_le_of_lt H2, 
     apply lt_succ_of_le, apply nat.le_add_left}
end

lemma sIns_fig_of_neg_of_moment
    {X : list B} {i : ℕ}
    (H1 : 1 ≤ length X)
    (H2 : ¬ (ρ X) - (ρ sDel X i) ≤ num_Is (sDel X i)) :
    X = sIns (sDel X i) i I :=
begin
cases moment_sub_sDel_total i H1,
    {apply absurd h.left, apply ne_of_gt,
     apply lt_of_le_of_lt,
        {apply num_RIs_le_sDel},
        {apply lt_of_not_ge H2}},
    {apply h.right}
end

lemma moment_sub_sDel_of_pos
    {X : list B} {i : ℕ}
    (H1 : 1 ≤ length X) 
    (H2 : (ρ X) - (ρ sDel X i) ≤ num_Is (sDel X i)) :
    (ρ X) - (ρ sDel X i) = num_RIs X i :=
begin
cases moment_sub_sDel_total i H1,
    {apply h.left},
    {apply absurd h.left, apply ne_of_lt,
     apply lt_of_le_of_lt H2, 
     apply lt_succ_of_le, apply nat.le_add_left}
end


lemma moment_sub_sDel_of_neg
    {X : list B} {i : ℕ}
    (H1 : 1 ≤ length X)
    (H2 : ¬ (ρ X) - (ρ sDel X i) ≤ num_Is (sDel X i)) :
    (ρ X) - (ρ sDel X i) = num_LOs X i + num_Is (sDel X i) + 1 :=
begin
cases moment_sub_sDel_total i H1,
    {apply absurd h.left, apply ne_of_gt,
     apply lt_of_le_of_lt,
        {apply num_RIs_le_sDel},
        {apply lt_of_not_ge H2}},
    {apply h.left}
end

end list