import .InsDel
import .B

open list nat

namespace list

def num_Os : list B → ℕ
| []     := 0
| (O::X) := num_Os X + 1 
| (I::X) := num_Os X

def num_Is : list B → ℕ
| []     := 0
| (O::X) := num_Is X  
| (I::X) := num_Is X + 1



-- num_Os 

lemma num_Os_cons
    (x : B) (X : list B) :
    num_Os (x::X) = num_Os [x] + num_Os X :=
begin
cases x,
    {unfold num_Os, rw add_comm},
    {unfold num_Os, rw zero_add}
end

lemma num_Os_le_cons
    (x : B) (X : list B) :
    num_Os X ≤ num_Os (x::X) :=
begin
cases x,
    {unfold num_Os, apply le_succ},
    {trivial}
end

lemma num_Os_cons_le
    (x : B) (X : list B) :
    num_Os (x::X) ≤ num_Os X + 1 :=
begin
cases x,
    {trivial},
    {unfold num_Os, apply le_succ}
end

lemma num_Os_sIns_zero
    (X : list B) (i : ℕ) :
    num_Os (sIns X i O) = num_Os X + 1 := 
begin
revert X, induction i with k IHi,
    {intro X, cases X with x X',
        {trivial},
        {trivial}},
    {intro X, cases X with x X',
        {trivial},
        {rw sIns, cases x,
            {unfold num_Os, rw (IHi X')},
            {unfold num_Os, rw (IHi X')}}}
end

lemma num_Os_sIns_one
    (X : list B) (i : ℕ) :
    num_Os (sIns X i I) = num_Os X := 
begin
revert X, induction i with k IHi,
    {intro X, cases X with x X',
        {trivial},
        {trivial}},
    {intro X, cases X with x X',
        {trivial},
        {rw sIns, cases x,
            {unfold num_Os, rw (IHi X')},
            {unfold num_Os, rw (IHi X')}}}
end

lemma num_Os_le_sIns 
    (X : list B) (i : ℕ) (a : B) :
    num_Os X ≤ num_Os (sIns X i a) :=
begin
cases a,
    {rw num_Os_sIns_zero, apply le_succ},
    {rw num_Os_sIns_one}
end

lemma num_Os_le_length 
    (X : list B) :
    num_Os X ≤ length X :=
begin
induction X with x X' IHX,
    {trivial},
    {cases x,
        {unfold num_Os, apply succ_le_succ IHX},
        {unfold num_Os, apply le_succ_of_le IHX}}
end

lemma num_Os_sDel_le_num_Os
    (X : list B) (i : ℕ) :
    num_Os (sDel X i) ≤ num_Os X :=
begin
revert i, induction X with x X' IHX,
    {intro i, trivial},
    {intro i, cases X' with x' X'',
        {rw sDel_singleton, apply nat.zero_le},
        {cases i,
            {rw sDel_zero, apply num_Os_le_cons},
            {unfold sDel, cases x,
                {unfold num_Os, apply succ_le_succ, apply IHX},
                {unfold num_Os, apply IHX}}}}
end

lemma num_Os_sDel_le_length
    (X : list B) (i : ℕ) :
    num_Os (sDel X i) ≤ length X - 1 :=
begin
apply le_trans,
    {apply num_Os_le_length},
    {apply length_sDel_le}
end

lemma num_Os_sDel_lt
    {X : list B} (i : ℕ) 
    (H : 1 ≤ length X ) :
    num_Os (sDel X i) < length X :=
begin
apply lt_of_le_of_lt,
    {apply num_Os_sDel_le_length},
    {apply nat.sub_lt,
        {apply lt_of_succ_le H},
        {apply lt_of_succ_le (le_refl 1)}}
end

-- num_Is

lemma num_Is_cons
    (x : B) (X : list B) :
    num_Is (x::X) = num_Is [x] + num_Is X :=
begin
cases x,
    {unfold num_Is, rw zero_add},
    {unfold num_Is, rw add_comm}
end

lemma num_Is_le_cons 
    (x : B) (X : list B) :
    num_Is X ≤ num_Is (x::X) :=
begin
cases x,
    {trivial},
    {unfold num_Is, apply le_succ}
end

lemma num_Is_cons_le 
    (x : B) (X : list B) :
    num_Is (x::X) ≤ num_Is X + 1 :=
begin
cases x,
    {unfold num_Is, apply le_succ},
    {trivial}
end

lemma num_Is_sDel
    (X : list B) (i : ℕ) :
    num_Is (sDel X i) ≤ num_Is X :=
begin
revert X, induction i with k IHi,
    {intro X, cases X with x X',
        {trivial},
        {rw sDel_zero, apply num_Is_le_cons}},
    {intro X, cases X with x X',
        {trivial},
        {cases X' with x' X'',
            {rw sDel_singleton, apply nat.zero_le},
            {unfold sDel, cases x,
                {unfold num_Is, apply IHi},
                {unfold num_Is, apply succ_le_succ, apply IHi}}}}
end

lemma num_Is_sIns_zero
    (X : list B) (i : ℕ) :
    num_Is (sIns X i O) = num_Is X := 
begin
revert X, induction i with k IHi,
    {intro X, cases X with x X',
        {trivial},
        {trivial}},
    {intro X, cases X with x X',
        {trivial},
        {rw sIns, cases x,
            {unfold num_Is, rw (IHi X')},
            {unfold num_Is, rw (IHi X')}}}
end

lemma num_Is_sIns_one
    (X : list B) (i : ℕ) :
    num_Is (sIns X i I) = num_Is X + 1:= 
begin
revert X, induction i with k IHi,
    {intro X, cases X with x X',
        {trivial},
        {trivial}},
    {intro X, cases X with x X',
        {trivial},
        {rw sIns, cases x,
            {unfold num_Is, rw (IHi X')},
            {unfold num_Is, rw (IHi X')}}}
end

lemma num_Is_le_sIns 
    (X : list B) (i : ℕ) (a : B) :
    num_Is X ≤ num_Is (sIns X i a) :=
begin
cases a,
    {rw num_Is_sIns_zero},
    {rw num_Is_sIns_one, apply le_succ}
end

lemma num_Is_sDel_zero
    {X : list B} {i : ℕ}
    (H : X = sIns (sDel X i) i O) :
    num_Is (sDel X i) = num_Is X :=
begin
rw H, rw sDel_sIns_id,
rw num_Is_sIns_zero
end

lemma num_Is_sDel_one
    {X : list B} {i : ℕ}
    (H : X = sIns (sDel X i) i I) :
    num_Is (sDel X i) = num_Is X - 1:=
begin
rw H, rw sDel_sIns_id,
rw num_Is_sIns_one,
rw nat.add_sub_cancel
end

lemma num_Is_sDel_le
    (X : list B) (i : ℕ) :
    num_Is (sDel X i) ≤ num_Is X :=
begin
revert i, induction X with x X' IHX,
    {intro i, trivial},
    {intro i, cases X' with x' X'',
        {rw sDel_singleton, apply nat.zero_le},
        {cases i,
            {rw sDel_zero, apply num_Is_le_cons},
            {unfold sDel, cases x,
                {unfold num_Is, apply IHX},
                {unfold num_Is, apply succ_le_succ, apply IHX}}}}
end

lemma le_num_Is_sDel
    (X : list B) (i : ℕ) :
    num_Is X - 1 ≤ num_Is (sDel X i) :=
begin
revert i, induction X with x X' IHX,
    {intro i, trivial},
    {intro i, cases X' with x' X'',
        {cases x,
            {trivial},
            {trivial}},
        {cases i,
            {cases x,
                {unfold num_Is, rw sDel_zero, apply nat.sub_le},
                {unfold num_Is, rw sDel_zero, rw nat.add_sub_cancel}},
            {unfold sDel, cases x,
                {unfold num_Is, apply IHX},
                {unfold num_Is, rw nat.add_sub_cancel,
                 cases num_Is (x' :: X''),
                    {apply nat.zero_le},
                    {apply succ_le_succ, apply IHX}}}}}
end

lemma num_Is_sDel_one_lt
    {X : list B} {i : ℕ}
    (H : X = sIns (sDel X i) i I) :
    num_Is (sDel X i) < num_Is X :=
begin
rw H, rw sDel_sIns_id,
rw num_Is_sIns_one,
apply lt_succ_of_le (le_refl (num_Is (sDel X i)))
end

lemma num_Is_le_length 
    (X : list B):
    num_Is X ≤ length X :=
begin
induction X with x X' IHX,
    {trivial},
    {cases x,
        {unfold num_Is, apply le_succ_of_le IHX},
        {unfold num_Is, apply succ_le_succ IHX}}
end

lemma num_Is_sDel_le_length
    (X : list B) (i : ℕ) :
    num_Is (sDel X i) ≤ length X - 1 :=
begin
apply le_trans,
    {apply num_Is_le_length},
    {apply length_sDel_le}  
end

lemma num_Is_sDel_lt
    {X : list B} (i : ℕ) 
    (H : 1 ≤ length X ) :
    num_Is (sDel X i) < length X :=
begin
cases X with x X',
    {apply false.elim (not_succ_le_zero 0 H)},
    {apply lt_of_le_of_lt,
        {apply num_Is_sDel_le_length},
        {rw length, rw nat.add_sub_cancel, apply lt_succ_self}}
end


-- num_Os_and_num_Is

lemma num_Os_add_num_Is
    (X : list B) :
    num_Os X + num_Is X = length X :=
begin
induction X with x X' IHX,
    {trivial},
    {cases x,
        {unfold num_Os, unfold num_Is, unfold length,
         rw add_right_comm, rw IHX},
        {unfold num_Os, unfold num_Is, unfold length,
         rw ← add_assoc, rw IHX}}
end

def num_LOs : list B → ℕ → ℕ 
| []     _       := 0
| (O::X) 0       := 0
| (O::X) (n + 1) := num_LOs X n + 1
| (I::X) 0       := 0
| (I::X) (n + 1) := num_LOs X n

def num_LIs : list B → ℕ → ℕ 
| []     _       := 0
| (O::X) 0       := 0
| (O::X) (n + 1) := num_LIs X n  
| (I::X) 0       := 0
| (I::X) (n + 1) := num_LIs X n + 1 



-- num_LOs

lemma num_LOs_zero
    (X : list B) :
    num_LOs X 0 = 0 :=
begin
cases X with x X',
    {trivial},
    {cases x,
        {trivial},
        {trivial}}
end

lemma num_LOs_le 
    (X : list B) (i : ℕ) :
    num_LOs X i ≤ i :=
begin
revert X, induction i with k IHi,
    {intro X, rw num_LOs_zero},
    {intro X, cases X with x X',
        {apply nat.zero_le},
        {cases x,
            {unfold num_LOs, apply succ_le_succ (IHi X')},
            {unfold num_LOs, apply le_succ_of_le (IHi X')}}}
end

lemma num_LOs_le_num_Os 
    (X : list B) (i : ℕ) :
    num_LOs X i ≤ num_Os X :=
begin
revert X, induction i with k IHi,
    {intro X, rw num_LOs_zero, apply nat.zero_le},
    {intro X, cases X with x X',
        {apply nat.zero_le},
        {cases x,
            {unfold num_LOs, unfold num_Os, apply succ_le_succ (IHi X')},
            {unfold num_LOs, unfold num_Os, apply IHi X'}}}
end

lemma num_LOs_of_ovrlen
    {X : list B} {i : ℕ} (H : length X ≤ i) :
    num_LOs X i = num_Os X :=
begin
revert i, induction X with x X' IHX,
    {intros i H, trivial},
    {intros i H, cases i,
        {apply false.elim (not_succ_le_zero (length X') H)},
        {cases x,
            {unfold num_LOs, unfold num_Os,
             rw IHX (le_of_succ_le_succ H)},
            {unfold num_LOs, unfold num_Os, 
             rw IHX (le_of_succ_le_succ H)}}}
end

lemma num_LOs_le_sDel
    {X : list B} {i : ℕ}
    (H : i + 1 ≤ length X) :
    num_LOs X i ≤ num_Os (sDel X i) :=
begin
revert X, induction i with k IHi,
    {intros X H, rw num_LOs_zero, apply nat.zero_le},
    {intros X H, cases X with x X',
        {apply nat.zero_le},
        {cases X' with x' X'',
            {apply absurd H, unfold length, 
             apply not_le_of_gt (succ_lt_succ (zero_lt_succ k))},
            {cases x,
                {unfold num_LOs, unfold sDel, unfold num_Os,
                 apply succ_le_succ (IHi (le_of_succ_le_succ H))},
                {unfold num_LOs, unfold sDel, unfold num_Os,
                 apply IHi (le_of_succ_le_succ H)}}}}
end

lemma num_LOs_le_length
    {X : list B} {i : ℕ}
    (H : i + 1 ≤ length X ):
    num_LOs X i ≤ length X - 1 :=
begin
apply le_trans,
    {apply num_LOs_le_sDel H},
    {apply le_trans,
        {apply num_Os_le_length},
        {apply length_sDel_le}}
end

lemma num_LOs_sIns
    {X : list B} {i : ℕ} (a : B) (H : i ≤ length X) :
    num_LOs (sIns X i a) i = num_LOs X i :=
begin
revert X, induction i with k IHi,
    {intros X H, repeat {rw num_LOs_zero}},
    {intros X H, cases X with x X',
        {apply false.elim (not_succ_le_zero k H)},
        {unfold sIns, cases x,
            {unfold num_LOs, rw IHi (le_of_succ_le_succ H)},
            {unfold num_LOs, rw IHi (le_of_succ_le_succ H)}}}
end

lemma num_LOs_sIns_one
    {i : ℕ} {X : list B} (H : length X ≤ i):
    num_LOs (sIns X i I) i = num_Os X :=
begin
revert X, induction i with k IHi,
    {intros X H, 
     have h : X = [], 
      from eq_nil_of_length_eq_zero (eq_zero_of_le_zero H),
     rw h, trivial},
    {intros X H, cases X with x X',
        {trivial},
        {unfold sIns, cases x,
            {unfold num_Os, unfold num_LOs, 
             rw IHi (le_of_succ_le_succ H)},
            {unfold num_Os, unfold num_LOs, 
             rw IHi (le_of_succ_le_succ H)}}}
end

lemma head_of_num_LOs
    {x : B} {X : list B} {i : ℕ}
    (H : 0 = num_LOs (x::X) (succ i)) :
    x = I :=
begin
cases x,
    {apply absurd H, apply ne_of_lt, 
     unfold num_LOs, apply lt_succ_of_le (nat.zero_le (num_LOs X i))},
    {refl}
end

--num_LIs 

lemma num_LIs_zero
    (X : list B):
    num_LIs X 0 = 0 :=
begin
cases X with x X',
    {trivial},
    {cases x,
        {trivial},
        {trivial}}
end

lemma num_LIs_le 
    (X : list B) (i : ℕ) :
    num_LIs X i ≤ i :=
begin
revert X, induction i with k IHi,
    {intro X, rw num_LIs_zero},
    {intro X, cases X with x X',
        {apply nat.zero_le},
        {cases x,
            {unfold num_LIs, apply le_succ_of_le (IHi X')},
            {unfold num_LIs, apply succ_le_succ (IHi X')}}}
end

lemma num_LIs_le_num_Is 
    (X : list B) (i : ℕ) :
    num_LIs X i ≤ num_Is X :=
begin
revert X, induction i with k IHi,
    {intro X, rw num_LIs_zero, apply nat.zero_le},
    {intro X, cases X with x X',
        {apply nat.zero_le},
        {cases x,
            {unfold num_LIs, unfold num_Is, apply IHi},
            {unfold num_LIs, apply succ_le_succ (IHi X')}}}
end

lemma num_LIs_of_ovrlen
    {X : list B} {i : ℕ} (H : length X ≤ i) :
    num_LIs X i = num_Is X :=
begin
revert i, induction X with x X' IHX,
    {intros i H, trivial},
    {intros i H, cases i,
        {apply false.elim (not_succ_le_zero (length X') H)},
        {cases x,
            {unfold num_LIs, unfold num_Is, rw IHX (le_of_succ_le_succ H)},
            {unfold num_LIs, unfold num_Is, rw IHX (le_of_succ_le_succ H)}}}
end

lemma num_LIs_le_sDel
    {X : list B} {i : ℕ}
    (H : i + 1 ≤ length X ):
    num_LIs X i ≤ num_Is (sDel X i) :=
begin
revert X, induction i with k IHi,
    {intros X H, rw num_LIs_zero, apply nat.zero_le},
    {intros X H, cases X with x X',
        {apply nat.zero_le},
        {cases X' with x' X'',
            {apply absurd H, unfold length, 
             apply not_le_of_gt (succ_lt_succ (zero_lt_succ k))},
            {cases x,
                {unfold num_LIs, unfold sDel, unfold num_Is,
                 apply IHi (le_of_succ_le_succ H)},
                {unfold num_LIs, unfold sDel, unfold num_Is,
                 apply succ_le_succ (IHi (le_of_succ_le_succ H))}}}}
end

lemma num_LIs_le_length
    {X : list B} {i : ℕ}
    (H : i + 1 ≤ length X ):
    num_LIs X i ≤ length X - 1 :=
begin
apply le_trans,
    {apply num_LIs_le_sDel H},
    {apply le_trans,
        {apply num_Is_le_length},
        {apply length_sDel_le}}
end

lemma num_LIs_sIns
    {X : list B} {i : ℕ} (a : B) (H : i ≤ length X) :
    num_LIs (sIns X i a) i = num_LIs X i :=
begin
revert X, induction i with k IHi,
    {intros X H, repeat {rw num_LIs_zero}},
    {intros X H, cases X with x X',
        {apply false.elim (not_succ_le_zero k H)},
        {unfold sIns, cases x,
            {unfold num_LIs, rw IHi (le_of_succ_le_succ H)},
            {unfold num_LIs, rw IHi (le_of_succ_le_succ H)}}}
end

lemma num_LIs_sIns_zero
    {i : ℕ} {X : list B} (H : length X ≤ i):
    num_LIs (sIns X i O) i = num_Is X :=
begin
revert X, induction i with k IHi,
    {intros X H, 
     have h : X = [], 
      from eq_nil_of_length_eq_zero (eq_zero_of_le_zero H),
     rw h, trivial},
    {intros X H, cases X with x X',
        {trivial},
        {unfold sIns, cases x,
            {unfold num_Is, unfold num_LIs, 
             rw IHi (le_of_succ_le_succ H)},
            {unfold num_Is, unfold num_LIs, 
             rw IHi (le_of_succ_le_succ H)}}}
end

lemma head_of_num_LIs
    {x : B} {X : list B} {i : ℕ}
    (H : 0 = num_LIs (x::X) (succ i)) :
    x = O :=
begin
cases x,
    {refl},
    {apply absurd H, apply ne_of_lt, 
     unfold num_LIs, apply lt_succ_of_le (nat.zero_le (num_LIs X i))},
end

lemma sIns_one_of_num_LOs_of_le
    {X : list B} {i j : ℕ} 
    (H1 : i ≤ j)
    (H2 : num_LOs (sIns X i I) i = num_LOs (sIns X j I) j) :
    sIns X i I = sIns X j I :=
begin
revert i j, induction X with x X' IHX,
    {intros i j H1 H2, trivial},
    {intros i j H1 H2, cases i,
        {cases j,
            {refl},
            {unfold sIns at H2, unfold num_LOs at H2,
             have h : x = I, 
             from head_of_num_LOs H2,
             rw ← sIns_one X' h,
             unfold sIns, rw cons_inj', 
             apply IHX (nat.zero_le j) _ , 
             rw sIns_zero, unfold num_LOs, 
             rw h at H2, unfold num_LOs at H2, apply H2}},
        {cases j,
            {apply false.elim (not_succ_le_zero i H1)},
            {unfold sIns, rw cons_inj',
             apply  IHX (le_of_succ_le_succ H1) _ ,
             unfold sIns at H2, cases x,
                {unfold num_LOs at H2, apply succ_inj H2},
                {unfold num_LOs at H2, apply H2}}}}
end

lemma sIns_one_of_num_LOs
    {X : list B} {i j : ℕ} 
    (H : num_LOs (sIns X i I) i = num_LOs (sIns X j I) j) :
    sIns X i I = sIns X j I :=
begin
cases decidable.em (i ≤ j) with h_le h_nle,
    {apply sIns_one_of_num_LOs_of_le h_le H},
    {apply eq.symm,
     apply sIns_one_of_num_LOs_of_le,
        {apply le_of_lt (lt_of_not_ge h_nle)},
        {apply eq.symm H}}       
end

-- num_LOs_and_num_LIs

lemma num_LOs_add_num_LIs
    {X : list B} {i : ℕ} (H : i ≤ length X) :
    num_LOs X i + num_LIs X i = i :=
begin
revert i, induction X with x X' IHX,
    {intros i H, 
     have h : i = 0, from eq_zero_of_le_zero H,
     rw h, trivial},
    {intros i H, cases i,
        {rw num_LOs_zero, rw num_LIs_zero},
        {cases x,
            {unfold num_LOs, unfold num_LIs, 
             rw add_right_comm, rw IHX (le_of_succ_le_succ H)},
            {unfold num_LOs, unfold num_LIs,
             rw ← add_assoc, rw IHX (le_of_succ_le_succ H)}}}
end

def min_num_LOs : list B → ℕ → ℕ 
|[]     _       := 0
|(O::X) 0       := 0
|(O::X) (n + 1) := min_num_LOs X n + 1
|(I::X) 0       := 0
|(I::X) (n + 1) := min_num_LOs X (n + 1) + 1

def min_num_LIs : list B → ℕ → ℕ 
|[]     _       := 0
|(O::X) 0       := 0
|(O::X) (n + 1) := min_num_LIs X (n + 1) + 1
|(I::X) 0       := 0
|(I::X) (n + 1) := min_num_LIs X n + 1



-- min_mun_LOs

lemma min_num_LOs_zero 
    (X : list B):
    min_num_LOs X 0 = 0 :=
begin
cases X with x X',
    {trivial},
    {cases x,
        {trivial},
        {trivial}}
end

lemma min_num_LOs_of_num_Os
    {X : list B} {i : ℕ} (H : num_Os X + 1 ≤ i) :
    min_num_LOs X i = length X :=
begin
revert i, induction X with x X' IHX,
    {intros i H, trivial},
    {intros i H, cases i,
        {apply false.elim (not_succ_le_zero (num_Os (x::X')) H)},
        {cases x,
            {unfold min_num_LOs, unfold length, rw IHX,
             apply le_of_succ_le_succ H},
            {unfold min_num_LOs, unfold length, rw IHX, apply H}}}
end

lemma min_num_LOs_le
    {X : list B} {i l : ℕ} (H : num_LOs X i = l) :
    min_num_LOs X l ≤ i :=
begin
revert i l, induction X with x X' IHX,
    {intros i l H, unfold min_num_LOs, apply nat.zero_le},
    {intros i l H, cases i,
        {rw num_LOs_zero at H, rw ← H,
         rw min_num_LOs_zero},
        {cases l,
            {rw min_num_LOs_zero, apply nat.zero_le},
            {cases x,
                {unfold num_LOs at H, unfold min_num_LOs,
                 apply succ_le_succ (IHX (succ_inj H))},
                {unfold num_LOs at H, unfold min_num_LOs,
                 apply succ_le_succ (IHX H)}}}}
end

lemma num_LOs_min_num_LOs_id
    {X : list B} {i : ℕ} (H : i ≤ num_Os X) :
    num_LOs X (min_num_LOs X i) = i :=
begin
revert i, induction X with x X' IHX,
    {intros i H, cases i,
        {trivial}, 
        {apply false.elim (not_succ_le_zero i H)}},
    {intros i H, cases i,
        {rw min_num_LOs_zero, rw num_LOs_zero},
        {cases x,
            {unfold min_num_LOs, unfold num_LOs, rw IHX,
             unfold num_Os at H, apply le_of_succ_le_succ H},
            {unfold min_num_LOs, unfold num_LOs, rw IHX, 
             unfold num_Os at H, apply H}}}
end

lemma min_num_LOs_le_length
    (X : list B) (i : ℕ) :
    min_num_LOs X i ≤ length X :=
begin
revert i, induction X with x X' IHX,
    {intro i, trivial},
    {intro i, cases i,
        {rw min_num_LOs_zero, apply nat.zero_le},
        {cases x,
            {unfold min_num_LOs, apply succ_le_succ (IHX i)},
            {unfold min_num_LOs, apply succ_le_succ (IHX (i + 1))}}}
end

lemma num_LOs_sIns_min_num_LOs_id
    {i : ℕ} {X : list B} (a : B) (H : i ≤ num_Os X):
    num_LOs (sIns X (min_num_LOs X i) a) 
    (min_num_LOs X i) = i :=
begin
rw num_LOs_sIns a (min_num_LOs_le_length X i), 
apply num_LOs_min_num_LOs_id H
end


-- min_num_LIs 

lemma min_num_LIs_zero
    (X : list B) :
    min_num_LIs X 0 = 0 :=
begin
cases X with x X',
    {trivial},
    {cases x,
        {trivial},
        {trivial}}
end

lemma min_num_LIs_of_num_Is
    {X : list B} {i : ℕ} (H : num_Is X + 1 ≤ i) :
    min_num_LIs X i = length X :=
begin
revert i, induction X with x X' IHX,
    {intros i H, trivial},
    {intros i H, cases i,
        {apply false.elim (not_succ_le_zero (num_Is (x::X')) H)},
        {cases x,
            {unfold min_num_LIs, unfold length, rw IHX, apply H},
            {unfold min_num_LIs, unfold length, rw IHX,
             apply le_of_succ_le_succ H}}}
end

lemma min_num_LIs_le
    {X : list B} {i l : ℕ} (H : num_LIs X i = l) :
    min_num_LIs X l ≤ i :=
begin
revert i l, induction X with x X' IHX,
    {intros i l H, unfold min_num_LIs, apply nat.zero_le},
    {intros i l H, cases i,
        {rw num_LIs_zero at H, rw ← H,
         rw min_num_LIs_zero},
        {cases l,
            {rw min_num_LIs_zero, apply nat.zero_le},
            {cases x,
                {unfold num_LIs at H, unfold min_num_LIs,
                 apply succ_le_succ (IHX H)},
                {unfold num_LIs at H, unfold min_num_LIs,
                 apply succ_le_succ (IHX (succ_inj H))}}}}
end

lemma num_LIs_min_num_LIs_id 
    {X : list B} {i : ℕ} (H : i ≤ num_Is X):
    num_LIs X (min_num_LIs X i) = i :=
begin
revert i, induction X with x X' IHX,
    {intros i H, cases i,
        {trivial}, 
        {apply false.elim (not_succ_le_zero i H)}},
    {intros i H, cases i,
        {rw min_num_LIs_zero, rw num_LIs_zero},
        {cases x,
            {unfold min_num_LIs, unfold num_LIs, rw IHX,
             unfold num_Is at H, apply H},
            {unfold min_num_LIs, unfold num_LIs, rw IHX, 
             unfold num_Is at H, apply le_of_succ_le_succ H}}}
end

lemma min_num_LIs_le_length 
    (X : list B) (i : ℕ) :
    min_num_LIs X i ≤ length X :=
begin
revert i, induction X with x X' IHX,
    {intro i, trivial},
    {intro i, cases i,
        {rw min_num_LIs_zero, apply nat.zero_le},
        {cases x,
            {unfold min_num_LIs, apply succ_le_succ (IHX (i + 1))},
            {unfold min_num_LIs, apply succ_le_succ (IHX i)}}}
end

lemma num_LIs_sIns_min_num_LIs_id 
    {i : ℕ} {X : list B} (a : B) (H : i ≤ num_Is X):
    num_LIs (sIns X (min_num_LIs X i) a) 
    (min_num_LIs X i) = i :=
begin
rw num_LIs_sIns a  
 (min_num_LIs_le_length X i), 
apply num_LIs_min_num_LIs_id H
end



def num_ROs : list B → ℕ → ℕ 
| []     _       := 0
| (x::X) 0       := num_Os X
| (x::X) (n + 1) := num_ROs X n

def num_RIs : list B → ℕ → ℕ 
| []     _       := 0
| (x::X) 0       := num_Is X
| (x::X) (n + 1) := num_RIs X n



-- num_ROs

lemma num_ROs_singleton
    (x : B) (i : ℕ) :
    num_ROs [x] i = 0 :=
begin
cases i,
    {trivial},
    {trivial}
end

lemma num_ROs_of_ovrlen
    {X : list B} {i : ℕ} (H : length X ≤ i) :
    num_ROs X i = 0 :=
begin
revert i, induction X with x X' IHX,
    {intros i H, trivial},
    {intros i H, cases i,
        {apply false.elim (not_succ_le_zero (length X') H)},
        {unfold num_ROs, apply IHX (le_of_succ_le_succ H)}}
end

lemma num_ROs_sIns
    {X : list B} {i : ℕ} (a : B) (H : length X ≤ i) :
    num_ROs (sIns X i a) i = num_ROs X i :=
begin
revert i, induction X with x X' IHX,
    {intros i H, unfold sIns, rw num_ROs_singleton, trivial},
    {intros i H, cases i,
        {apply false.elim (not_succ_le_zero (length X') H)},
        {unfold num_ROs, rw sIns, rw num_ROs, 
         apply IHX (le_of_succ_le_succ H)}}
end

lemma num_ROs_le
    (X : list B) (i : ℕ) :
    num_ROs X i ≤ num_Os X :=
begin
revert i, induction X with x' X' IHX,
    {intro i, apply nat.zero_le},
    {intro i, cases i,
        {unfold num_ROs, apply num_Os_le_cons},
        {unfold num_ROs, 
         apply le_trans (IHX i) (num_Os_le_cons x' X')}}
end

lemma num_ROs_le_cons 
    (x : B) (X : list B) (i : ℕ) :
    num_ROs X i ≤ num_ROs (x::X) i :=
begin
revert x i, induction X with x' X' IHX,
    {intros x i, apply nat.zero_le},
    {intros x i, cases i,
        {unfold num_ROs, apply num_Os_le_cons},
        {unfold num_ROs, apply IHX}}
end

lemma num_ROs_le_tail 
    (x : B) (X : list B) (i : ℕ) :
    num_ROs (x::X) i ≤ num_Os X :=
begin
cases i with k IHi,
    {trivial},
    {unfold num_ROs, apply num_ROs_le}
end

lemma num_ROs_succ_le
    (X : list B) (i : ℕ) :
    num_ROs X (succ i) ≤ num_ROs X i :=
begin
revert i, induction X with x' X' IHX,
    {intro i, trivial},
    {intro i, cases i,
        {unfold num_ROs, apply num_ROs_le},
        {unfold num_ROs, apply IHX}}
end

lemma num_ROs_le_sDel
    (X : list B) (i : ℕ) :
    num_ROs X i ≤ num_Os (sDel X i) :=
begin
revert i, induction X with x X' IHX,
    {intro i, trivial},
    {intro i, cases X' with x' X'',
        {rw num_ROs_singleton, apply nat.zero_le},
        {cases i,
            {unfold sDel, unfold num_ROs},
            {unfold sDel, unfold num_ROs,
             apply le_trans (IHX i) 
              (num_Os_le_cons x (sDel (x'::X'') i))}}}
end

lemma num_ROs_le_length
    (X : list B) (i : ℕ) :
    num_ROs X i ≤ length X - (i + 1) :=
begin
revert i, induction X with x' X' IHX,
    {intro i, apply nat.zero_le},
    {intro i, cases i,
        {unfold num_ROs, unfold length, rw succ_sub_succ,
         apply num_Os_le_length},
        {unfold num_ROs, unfold length, rw succ_sub_succ, apply IHX}}
end

lemma head_of_num_ROs
    {x : B} {X : list B} {i : ℕ}
    (H : num_Os (x::X)  = num_ROs (sIns X i I) i) :
    x = I :=
begin
cases x,
    {apply absurd H, apply ne_of_gt, apply lt_succ_of_le,
     apply le_trans (num_ROs_le (sIns X i I) i),
     rw num_Os_sIns_one},
    {refl}
end

lemma num_Os_sDel_lt_sub_mod
    (X : list B) (i : ℕ) :
    num_Os (sDel X i) < num_ROs X i + i + 1 :=
begin
revert i, induction X with x X' IHX,
    {intro i, apply zero_lt_succ},
    {intro i, cases X' with x' X'',
        {rw sDel_singleton, apply zero_lt_succ},
        {cases i,
            {unfold sDel, unfold num_ROs, apply lt_succ_self},
            {unfold sDel, unfold num_ROs, 
             apply lt_of_le_of_lt 
              (num_Os_cons_le x (sDel (x'::X'') i))
              (succ_lt_succ (IHX i))}}}
end

-- num_RIs

lemma num_RIs_singleton
    (x : B) (i : ℕ) :
    num_RIs [x] i = 0 :=
begin
cases i,
    {trivial},
    {trivial}
end

lemma num_RIs_of_ovrlen
    {X : list B} {i : ℕ} (H : length X ≤ i) :
    num_RIs X i = 0 :=
begin
revert i, induction X with x X' IHX,
    {intros i H, trivial},
    {intros i H, cases i,
        {apply false.elim (not_succ_le_zero (length X') H)},
        {unfold num_RIs, apply IHX (le_of_succ_le_succ H)}}
end

lemma num_RIs_sIns
    {X : list B} {i : ℕ} (a : B) (H : length X ≤ i) :
    num_RIs (sIns X i a) i = num_RIs X i :=
begin
revert i, induction X with x X' IHX,
    {intros i H, unfold sIns, rw num_RIs_singleton, trivial},
    {intros i H, cases i,
        {apply false.elim (not_succ_le_zero (length X') H)},
        {unfold num_RIs, rw sIns, rw num_RIs, 
         apply IHX (le_of_succ_le_succ H)}}
end

lemma num_RIs_le
    (X : list B) (i : ℕ) :
    num_RIs X i ≤ num_Is X :=
begin
revert i, induction X with x' X' IHX,
    {intro i, apply nat.zero_le},
    {intro i, cases i,
        {unfold num_RIs, apply num_Is_le_cons},
        {unfold num_RIs, 
         apply le_trans (IHX i) (num_Is_le_cons x' X')}}
end

lemma num_RIs_le_sDel
    (X : list B) (i : ℕ) :
    num_RIs X i ≤ num_Is (sDel X i) :=
begin
revert i, induction X with x' X' IHX,
    {intro i, apply nat.zero_le},
    {intro i, cases X' with x' X'',
        {rw num_RIs_singleton, apply nat.zero_le},
        {cases i,
            {unfold num_RIs, unfold sDel},
            {unfold num_RIs, unfold sDel, apply le_trans,
                {apply IHX},
                {apply num_Is_le_cons}}}}
end

lemma num_RIs_le_cons
    (x : B) (X : list B) (i : ℕ) :
    num_RIs X i ≤ num_RIs (x::X) i :=
begin
revert x i, induction X with x' X' IHX,
    {intros x i, unfold num_RIs, apply nat.zero_le},
    {intros x i, cases i,
        {unfold num_RIs, apply num_Is_le_cons},
        {unfold num_RIs, apply IHX}}
end

lemma num_RIs_le_tail 
    (x : B) (X : list B) (i : ℕ) :
    num_RIs (x::X) i ≤ num_Is X :=
begin
cases i with k IHi,
    {trivial},
    {unfold num_RIs, apply num_RIs_le}
end

lemma num_RIs_succ_le
    (X : list B) (i : ℕ) :
    num_RIs X (succ i) ≤ num_RIs X i :=
begin
revert i, induction X with x' X' IHX,
    {intro i, trivial},
    {intro i, cases i,
        {unfold num_RIs, apply num_RIs_le},
        {unfold num_RIs, apply IHX}}
end

lemma num_RIs_le_length
    (X : list B) (i : ℕ) :
    num_RIs X i ≤ length X - (i + 1) :=
begin
revert i, induction X with x' X' IHX,
    {intro i, apply nat.zero_le},
    {intro i, cases i,
        {unfold num_RIs, unfold length, rw succ_sub_succ,
         apply num_Is_le_length},
        {unfold num_RIs, unfold length, rw succ_sub_succ, apply IHX}}
end

lemma head_of_num_RIs
    {x : B} {X : list B} {i : ℕ}
    (H : num_Is (x::X)  = num_RIs (sIns X i O) i) :
    x = O :=
begin
cases x,
    {refl},
    {apply absurd H, apply ne_of_gt, apply lt_succ_of_le,
     apply le_trans (num_RIs_le (sIns X i O) i),
     rw num_Is_sIns_zero}
end


lemma sIns_zero_of_num_RIs_of_le
    {X : list B} {i j : ℕ} 
    (H1 : i ≤ j)
    (H2 : num_RIs (sIns X i O) i = num_RIs (sIns X j O) j) :
    sIns X i O = sIns X j O :=
begin
revert i j, induction X with x X' IHX,
    {intros i j H1 H2, trivial},
    {intros i j H1 H2, cases i,
        {cases j,
            {refl},
            {unfold sIns at H2, unfold num_RIs at H2,
             have h : x = O, 
             from head_of_num_RIs H2,
             rw ← sIns_one X' h,
             unfold sIns, rw cons_inj',
             apply IHX (nat.zero_le j) _ , 
             rw sIns_zero, unfold num_RIs, 
             rw h at H2, unfold num_Is at H2, apply H2}},
        {cases j,
            {apply false.elim (not_succ_le_zero i H1)},
            {unfold sIns, rw cons_inj',
             apply IHX (le_of_succ_le_succ H1) _ ,
             unfold sIns at H2, unfold num_RIs at H2, apply H2}}}
end

lemma sIns_zero_of_num_RIs
    {X : list B} {i j : ℕ} 
    (H : num_RIs (sIns X i O) i = num_RIs (sIns X j O) j) :
    sIns X i O = sIns X j O :=
begin
cases decidable.em (i ≤ j) with h_le h_nle,
    {apply sIns_zero_of_num_RIs_of_le h_le H},
    {apply eq.symm,
     apply sIns_zero_of_num_RIs_of_le,
        {apply le_of_lt (lt_of_not_ge h_nle)},
        {apply eq.symm H}}       
end

lemma num_Is_sDel_lt_sub_mod
    (X : list B) (i : ℕ) :
    num_Is (sDel X i) < num_RIs X i + i + 1 :=
begin
revert i, induction X with x X' IHX,
    {intro i, apply zero_lt_succ},
    {intro i, cases X' with x' X'',
        {rw sDel_singleton, apply zero_lt_succ},
        {cases i,
            {unfold sDel, unfold num_RIs, apply lt_succ_self},
            {unfold sDel, unfold num_RIs, 
             apply lt_of_le_of_lt 
              (num_Is_cons_le x (sDel (x'::X'') i))
              (succ_lt_succ (IHX i))}}}
end
-- num_LOs_and_num_ROs

lemma num_LOs_add_num_ROs
    {X : list B} {i : ℕ} (H : i + 1 ≤ length X) :
    num_LOs X i + num_ROs X i = num_Os (sDel X i) :=
begin
revert i, induction X with x X' IHX,
    {intros i H, apply false.elim (not_succ_le_zero i H)},
    {intros i H, cases i,
        {rw num_LOs_zero, unfold num_ROs, rw nat.zero_add,
         rw sDel_zero},
        {cases X' with x' X'',
            {apply absurd H,
             apply not_le_of_gt (lt_succ_of_le (succ_le_succ (nat.zero_le i)))},
            {cases x,
                {unfold num_LOs, unfold num_ROs, rw add_right_comm,
                 unfold sDel, unfold num_Os, rw IHX (le_of_succ_le_succ H)},
                {unfold num_LOs, unfold num_ROs, 
                 unfold sDel, unfold num_Os, rw IHX (le_of_succ_le_succ H)}}}}
end

-- num_LIs_and_num_RIs

lemma num_LIs_add_num_RIs
    {X : list B} {i : ℕ} (H : i + 1 ≤ length X)  :
    num_LIs X i + num_RIs X i = num_Is (sDel X i) :=
begin
revert i, induction X with x X' IHX,
    {intros i H, apply false.elim (not_succ_le_zero i H)},
    {intros i H, cases i,
        {rw num_LIs_zero, unfold num_RIs, rw nat.zero_add,
         rw sDel_zero},
        {cases X' with x' X'',
            {apply absurd H, 
             apply not_le_of_gt (lt_succ_of_le (succ_le_succ (nat.zero_le i)))},
            {cases x,
                {unfold num_LIs, unfold num_RIs,
                 unfold sDel, unfold num_Is, rw IHX (le_of_succ_le_succ H)},
                {unfold num_LIs, unfold num_RIs, rw add_right_comm,
                 unfold sDel, unfold num_Is, rw IHX (le_of_succ_le_succ H)}}}}
end



def max_num_ROs : list B → ℕ → ℕ 
| []      _ := 0
| (x::X)  n := if num_Os X + 1 ≤ n 
               then max_num_ROs X n else max_num_ROs X n + 1

def max_num_RIs : list B → ℕ → ℕ 
| []     _ := 0
| (x::X) n := if num_Is X + 1 ≤ n 
              then max_num_RIs X n else max_num_RIs X n + 1


-- max_num_ROs

lemma max_num_ROs_zero
    (X : list B) :
    max_num_ROs X 0 = length X :=
begin
induction X with x X' IHX,
    {trivial},
    {unfold max_num_ROs, 
     cases decidable.em (num_Os X' + 1 ≤ 0) with h_le h_nle,
        {apply false.elim (not_succ_le_zero (num_Os X') h_le)},
        {rw if_neg h_nle, unfold length, rw IHX}}
end

lemma max_num_ROs_of_num_Os 
    {X : list B} {i : ℕ} (H : num_Os X + 1 ≤ i) :
    max_num_ROs X i = 0 :=
begin
revert i, induction X with x X' IHX,
    {intros i H, trivial},
    {intros i H, cases i,
        {apply false.elim (not_succ_le_zero (num_Os (x::X')) H)},
        {unfold max_num_ROs, 
         have h : num_Os X' + 1 ≤ succ i, 
         apply le_trans (nat.add_le_add_right (num_Os_le_cons x X') 1) H,
         rw if_pos h, rw IHX h}}
end

lemma le_max_num_ROs
    {X : list B} {i r : ℕ} (H1 : i + 1 ≤ length X) 
    (H2 : num_ROs X i = r) :
    i ≤ max_num_ROs X r :=
begin
revert i r, induction X with x X' IHX,
    {intros i r H1 H2, apply false.elim (not_succ_le_zero i H1)},
    {intros i r H1 H2, cases i,
        {apply nat.zero_le},
        {unfold max_num_ROs, 
         have h : ¬ num_Os X' + 1 ≤ r, 
         unfold num_ROs at H2, rw ← H2,
         apply not_le_of_gt (lt_succ_of_le (num_ROs_le X' i)), 
         rw if_neg h, apply succ_le_succ (IHX (le_of_succ_le_succ H1) _ ), 
         unfold num_ROs at H2, rw ← H2}}
end

lemma max_num_ROs_le_length 
    (X : list B) (i : ℕ) :
    max_num_ROs X i ≤ length X :=
begin
revert i, induction X with x X' IHX,
    {intro i, trivial},
    {intro i, cases i,
        {rw max_num_ROs_zero},
        {unfold max_num_ROs, 
         cases decidable.em (num_Os X' + 1 ≤ succ i) with h_le h_nle,
            {rw if_pos h_le, apply le_trans (IHX (succ i)) (le_succ (length X'))},
            {rw if_neg h_nle, apply succ_le_succ (IHX (i + 1))}}}
end

lemma num_ROs_sIns_max_num_ROs_id
    {X : list B} {i : ℕ} (a : B) (H : i  ≤ num_Os X) :
    num_ROs (sIns X (max_num_ROs X i) a) 
    (max_num_ROs X i) = i :=
begin
revert i, induction X with x X' IHX,
    {intros i H, cases i,
        {trivial},
        {apply false.elim (not_succ_le_zero i H)}},
    {intros i H, unfold max_num_ROs, 
     cases decidable.em (num_Os X' + 1 ≤ i) with h_le h_nle,
        {rw if_pos h_le, 
         have h : max_num_ROs X' i = 0, 
         apply max_num_ROs_of_num_Os h_le,
         rw h, unfold sIns, unfold num_ROs, 
         apply le_antisymm (le_trans (num_Os_cons_le x X') h_le) H},
        {rw if_neg h_nle, unfold sIns, unfold num_ROs, 
         rw IHX (le_of_lt_succ (lt_of_not_ge h_nle))}}
end


-- max_num_RIs


lemma max_num_RIs_zero
    (X : list B) :
    max_num_RIs X 0 = length X :=
begin
induction X with x X' IHX,
    {trivial},
    {unfold max_num_RIs, 
     cases decidable.em (num_Is X' + 1 ≤ 0) with h_le h_nle,
        {apply false.elim (not_succ_le_zero (num_Is X') h_le)},
        {rw if_neg h_nle, unfold length, rw IHX}}
end

lemma max_num_RIs_of_num_Is 
    {X : list B} {i : ℕ} (H : num_Is X + 1 ≤ i) :
    max_num_RIs X i = 0 :=
begin
revert i, induction X with x X' IHX,
    {intros i H, trivial},
    {intros i H, cases i,
        {apply false.elim (not_succ_le_zero (num_Is (x::X')) H)},
        {unfold max_num_RIs, 
         have h : num_Is X' + 1 ≤ succ i, 
         apply le_trans (nat.add_le_add_right (num_Is_le_cons x X') 1) H,
         rw if_pos h, rw IHX h}}
end

lemma le_max_num_RIs
    {X : list B} {i r : ℕ} (H1 : i + 1 ≤ length X) (H2 : num_RIs X i = r) :
    i ≤ max_num_RIs X r :=
begin
revert i r, induction X with x X' IHX,
    {intros i r H1 H2, apply false.elim (not_succ_le_zero i H1)},
    {intros i r H1 H2, cases i,
        {apply nat.zero_le},
        {unfold max_num_RIs, 
         have h : ¬ num_Is X' + 1 ≤ r, 
         unfold num_RIs at H2, rw ← H2,
         apply not_le_of_gt (lt_succ_of_le (num_RIs_le X' i)), 
         rw if_neg h, apply succ_le_succ (IHX (le_of_succ_le_succ H1) _), 
         unfold num_RIs at H2, rw ← H2}}
end

lemma max_num_RIs_le_length 
    (X : list B) (i : ℕ) :
    max_num_RIs X i ≤ length X :=
begin
revert i, induction X with x X' IHX,
    {intro i, trivial},
    {intro i, cases i,
        {rw max_num_RIs_zero},
        {unfold max_num_RIs, 
         cases decidable.em (num_Is X' + 1 ≤ succ i) with h_le h_nle,
            {rw if_pos h_le, apply le_trans (IHX (succ i)) (le_succ (length X'))},
            {rw if_neg h_nle, apply succ_le_succ (IHX (i + 1))}}}
end

lemma num_RIs_sIns_max_num_RIs_id
    {X : list B} {i : ℕ} (a : B) (H : i  ≤ num_Is X) :
    num_RIs (sIns X (max_num_RIs X i) a) 
    (max_num_RIs X i) = i :=
begin
revert i, induction X with x X' IHX,
    {intros i H, cases i,
        {trivial},
        {apply false.elim (not_succ_le_zero i H)}},
    {intros i H, unfold max_num_RIs, 
     cases decidable.em (num_Is X' + 1 ≤ i) with h_le h_nle,
        {rw if_pos h_le, 
         have h : max_num_RIs X' i = 0, 
         apply max_num_RIs_of_num_Is h_le,
         rw h, unfold sIns, unfold num_RIs, 
         apply le_antisymm (le_trans (num_Is_cons_le x X') h_le) H},
        {rw if_neg h_nle, unfold sIns, unfold num_RIs, 
         rw IHX (le_of_lt_succ (lt_of_not_ge h_nle))}}
end

lemma num_Is_repO {n : ℕ} :  
  num_Is (repeat O n) = 0 :=
begin
induction n with n IHn,
  {trivial},
  {unfold list.repeat, unfold num_Is, apply IHn}
end 

lemma rep0_of_num_Is0 (X : list B) (H : num_Is X = 0) :  
  X = repeat O (length X) :=
begin
induction X with x X' IHX,
  {trivial},
  {unfold list.length, unfold list.repeat, cases x,
    {congr, unfold num_Is at H, apply IHX H},
    {contradiction}}
end 

lemma sDel_of_num_Is0 (X : list B) (i : ℕ) (H : num_Is X = 0) :  
  sDel X i = repeat O (length X - 1) :=
by {rw rep0_of_num_Is0 X H, rw sDel_repeat, rw length_repeat}

lemma sDel_of_num_Is1 (X : list B) (H : num_Is X = 1) :  
  ∃ i : ℕ, sDel X i = repeat O (length X - 1) :=
begin
induction X with x X' IHX,
  {apply exists.intro 0, rw sDel_nil, refl},
  {cases X' with x' X'', 
    {apply exists.intro 0, refl},
    {cases x,
      {unfold num_Is at H, cases IHX H with i IHXi, 
       apply exists.intro (i + 1), unfold sDel, rw IHXi, 
       unfold list.length, repeat{rw nat.succ_sub_one}, refl},
      {apply exists.intro 0, unfold sDel, unfold num_Is at H,
       rw rep0_of_num_Is0 _ (nat.succ_inj H), simp}}}
end 

lemma sDel_of_le_num_Is1 (X : list B) (H : num_Is X ≤ 1) :  
  ∃ i : ℕ, sDel X i = repeat O (length X - 1) :=
begin
cases eq_or_lt_of_le H, 
  {apply sDel_of_num_Is1 _ h},
  {apply exists.intro 0, apply sDel_of_num_Is0,
   apply nat.eq_zero_of_le_zero, apply nat.le_of_lt_succ h}
end 

lemma sDel_ne_repO (X : list B) (H : ¬ num_Is X ≤ 1) :  
  ∀ i j : ℕ, sDel X j ≠ sDel (repeat O (length X)) i :=
begin
intros i j, assume h,
have h1 : 1 ≤ num_Is (sDel X j),
  {apply le_trans _,
    {apply le_num_Is_sDel},
    {apply nat.le_sub_left_of_add_le, 
     apply lt_of_not_ge H}},
have h2 : num_Is (sDel X j) = 0,
  {rw h, rw sDel_repeat, rw num_Is_repO},
rw h2 at h1, apply absurd h1,
apply not_le_of_lt, apply nat.lt_succ_of_le, 
apply nat.zero_le
end 


lemma num_Is_repI {n : ℕ} :  
  num_Is (repeat I n) = n :=
begin
induction n with n IHn,
  {trivial},
  {unfold list.repeat, unfold num_Is,
   rw nat.succ_inj', apply IHn}
end 

lemma rep1_of_num_Is_length (X : list B) (H : num_Is X = (length X)) :  
  X = repeat I (length X) :=
begin
induction X with x X' IHX,
  {trivial},
  {unfold list.length, unfold list.repeat, cases x,
    {unfold num_Is at H, apply absurd H, apply ne_of_lt,
     apply nat.lt_succ_of_le, apply num_Is_le_length},
    {congr, unfold num_Is at H, apply IHX (nat.succ_inj H)}}
end 

lemma sDel_of_num_Is_length (X : list B) (i : ℕ) (H : num_Is X = (length X)) :  
  sDel X i = repeat I (length X - 1) :=
by {rw rep1_of_num_Is_length X H, rw sDel_repeat, rw length_repeat}

lemma sDel_of_num_Is_length_sub (X : list B) (H : num_Is X = (length X) - 1) :  
  ∃ i : ℕ, sDel X i = repeat I (length X - 1) :=
begin
induction X with x X' IHX,
  {apply exists.intro 0, trivial},
  {cases X' with x' X'', 
    {apply exists.intro 0, trivial},
    {cases x,
      {apply exists.intro 0, unfold sDel, unfold num_Is at H,
       rw rep1_of_num_Is_length _ H, simp},
      {cases IHX (nat.succ_inj H) with i IHXi, 
        {apply exists.intro (i + 1),
         unfold sDel, rw IHXi, unfold list.length, 
         repeat{rw nat.succ_sub_one}, unfold list.repeat}}}}
end 

lemma sDel_of_le_num_Is_length_sub (X : list B) (H : num_Is X ≥ (length X) - 1) :  
  ∃ i : ℕ, sDel X i = repeat I (length X - 1) :=
begin
cases eq_or_lt_of_le H,
  {apply sDel_of_num_Is_length_sub X h.symm},
  {apply exists.intro 0,
   have h : num_Is X = length X, 
    {apply le_antisymm, 
      {apply num_Is_le_length},
      {apply nat.le_of_pred_lt, apply h}},
   apply sDel_of_num_Is_length X 0 h}
end 

lemma sDel_ne_repI (X : list B) (H : ¬ num_Is X ≥ (length X) - 1) :  
  ∀ i j : ℕ, sDel X j ≠ sDel (repeat I (length X)) i :=
begin
intros i j, assume h,
have h1 : num_Is (sDel X j) < (length X) - 1,
  {apply lt_of_le_of_lt, 
    {apply num_Is_sDel_le},
    {apply lt_of_not_ge H}},
have h2 : num_Is (sDel X j) = (length X) - 1,
  {rw h, rw sDel_repeat, rw num_Is_repI},
rw h2 at h1, apply absurd h1,
apply not_lt_of_ge, apply le_refl
end 

end list

namespace vector

def wt {n : ℕ} (X : vector B n) : ℕ :=
    list.num_Is (to_list X)

lemma wt_repO {n : ℕ} :  
  wt (repeat O n) = 0 :=
by {unfold vector.repeat, unfold wt, apply list.num_Is_repO}

lemma rep_zero_of_wt0 {n : ℕ} (X : vector B n) (H : wt X = 0) :  
  X = repeat O n :=
begin
rw subtype.ext, repeat {rw ← to_list},
unfold repeat, rw to_list_mk, 
rw list.rep0_of_num_Is0 _ H, rw to_list_length
end 

lemma sDel_of_wt0 {n : ℕ} (X : vector B n) (i : ℕ) (H : wt X = 0) :  
  sDel X i = repeat O (n - 1) :=
by {rw rep_zero_of_wt0 X H, rw sDel_repeat}

lemma sDel_of_wt1 {n : ℕ} (X : vector B n) (H : wt X = 1) :  
  ∃ i : ℕ, sDel X i = repeat O (n - 1) :=
begin
cases list.sDel_of_num_Is1 (to_list X) H with i h_i,
apply exists.intro i, 
rw subtype.ext, unfold sDel, unfold repeat, 
repeat {rw ← to_list, rw to_list_mk}, 
rw h_i, rw to_list_length
end 

lemma sDel_of_le_wt1 {n : ℕ} (X : vector B n) (H : wt X ≤ 1) :  
  ∃ i : ℕ, sDel X i = repeat O (n - 1) :=
begin
cases eq_or_lt_of_le H, 
  {apply sDel_of_wt1 _ h},
  {apply exists.intro 0, apply sDel_of_wt0,
   apply nat.eq_zero_of_le_zero, apply nat.le_of_lt_succ h}
end 

lemma sDel_ne_repO {n : ℕ} (X : vector B n) (H : ¬ wt X ≤ 1) :  
  ∀ i j : ℕ, sDel X j ≠ sDel (repeat O n) i :=
begin
intros i j, assume h,
have h1 : 1 ≤ wt (sDel X j),
  {apply le_trans _,
    {apply list.le_num_Is_sDel},
    {apply nat.le_sub_left_of_add_le, 
     apply lt_of_not_ge H}},
have h2 : wt (sDel X j) = 0,
  {rw h, rw sDel_repeat, rw wt_repO},
rw h2 at h1, apply absurd h1,
apply not_le_of_lt, apply nat.lt_succ_of_le, 
apply nat.zero_le
end 

lemma wt_repI {n : ℕ} :  
  wt (repeat I n) = n :=
by {unfold vector.repeat, unfold wt, apply list.num_Is_repI}

lemma rep1_of_wt_length {n : ℕ} (X : vector B n) (H : wt X = n) :  
  X = repeat I n :=
begin
rw subtype.ext, repeat {rw ← to_list},
unfold repeat, rw to_list_mk, 
rw list.rep1_of_num_Is_length (to_list X),
    {rw to_list_length},
    {rw to_list_length, apply H}
end 

lemma sDel_of_wt_length {n : ℕ} (X : vector B n) (i : ℕ) (H : wt X = n) :  
  sDel X i = repeat I (n - 1) :=
by {rw rep1_of_wt_length X H, rw sDel_repeat}

lemma sDel_of_wt_length_sub {n : ℕ} (X : vector B n) (H : wt X = n - 1) :  
  ∃ i : ℕ, sDel X i = repeat I (n - 1) :=
begin
cases list.sDel_of_num_Is_length_sub (to_list X) _ with i h_i,
    {apply exists.intro i, 
     rw subtype.ext, unfold sDel, unfold repeat, 
     repeat {rw ← to_list, rw to_list_mk}, 
     rw h_i, rw to_list_length},
    {rw to_list_length, apply H}
end 

lemma sDel_of_le_wt_length_sub {n : ℕ} (X : vector B n) (H : wt X ≥ n - 1) :  
  ∃ i : ℕ, sDel X i = repeat I (n - 1) :=
begin
cases eq_or_lt_of_le H,
  {apply sDel_of_wt_length_sub X h.symm},
  {apply exists.intro 0,
   have h : wt X = n, 
    {apply le_antisymm, 
      {unfold wt, apply le_trans,
        {apply list.num_Is_le_length},
        {rw to_list_length}},
      {apply nat.le_of_pred_lt, apply h}},
   apply sDel_of_wt_length X 0 h}
end 

lemma sDel_ne_repI {n : ℕ} (X : vector B n) (H : ¬ wt X ≥ n - 1) :  
  ∀ i j : ℕ, sDel X j ≠ sDel (repeat I n) i :=
begin
intros i j, assume h,
have h1 : wt (sDel X j) < n - 1,
  {apply lt_of_le_of_lt, 
    {apply list.num_Is_sDel_le},
    {apply lt_of_not_ge H}},
have h2 : wt (sDel X j) = n - 1,
  {rw h, rw sDel_repeat, rw wt_repI},
rw h2 at h1, apply absurd h1,
apply not_lt_of_ge, apply le_refl
end 

end vector