import data.vector data.finset
import .B .InsDel .DelCode

variable {α : Type}
variable [decidable_eq α] 

open vector 

def vector.repr {α : Type} [has_repr α] {n : ℕ} : 
  vector α n → string :=
by {intro v, apply list.repr (to_list v)}

instance {α : Type} [has_repr α] {n : ℕ} : 
  has_repr (vector α n) :=
⟨vector.repr⟩

namespace finset 

def is_optimal {n : ℕ} (C : finset (vector B n)) : Prop := 
  is_DelCode C ∧ ∀ s ∈ @DelCodes n, card s ≤ card C

instance is_optimal_decidable {n : ℕ} : 
  decidable_pred (@is_optimal n) :=
by {intro s, unfold is_optimal, apply_instance}

def is_optimal_Area {n : ℕ} (C : finset (vector B n)) : Prop := 
  is_DelCode C ∧ ∀ s ∈ @DelCodes_Area n, card s ≤ card C

instance is_optimal_Area_decidable {n : ℕ} : 
  decidable_pred (@is_optimal_Area n) :=
by {intro s, unfold is_optimal_Area, apply_instance}

lemma optimal_Area_iff_optimal {n : ℕ} (C : finset (vector B n)) :
  is_optimal_Area C ↔ is_optimal C :=
begin
unfold is_optimal_Area, unfold is_optimal,
unfold DelCodes_Area, rw DelCodes_Area3_eq_DelCodes
end

def is_repO_in {n : ℕ} (C : finset (vector B n)) : Prop :=
  repeat O n ∈ C

instance decidable_pred_is_repO_in {n : ℕ} : 
  decidable_pred (@is_repO_in n) :=
by {intro s, unfold is_repO_in, apply_instance}

lemma repO_notin {n : ℕ} (C : finset (vector B n)):
  repeat O n ∉ filter (not wt_le) C :=
begin
assume h, rw mem_filter at h, 
have hle : wt_le (repeat O n), 
  {unfold wt_le, rw wt_repO, apply nat.zero_le},
apply absurd hle h.right
end

def insert_repO {n : ℕ} (C : finset (vector B n)) : finset (vector B n) := 
  insert (repeat O n) (filter (not wt_le) C)

lemma DelCode_insert_repO
  {n : ℕ} (C : finset (vector B n)) (H : is_DelCode C) :
  is_DelCode (insert_repO C) :=
begin
unfold insert_repO, apply DelCode_insert,
  {apply DelCode_filter C H},
  {intros y h, rw dS_inter_dS_empty_iff,
   intros i j, apply ne.symm, apply sDel_ne_repO,
   rw mem_filter at h, apply h.right}
end

lemma card_insert_repO
  {n : ℕ} (C : finset (vector B n)) (H : is_DelCode C) :
  card C ≤ card (insert_repO C) :=
begin
unfold insert_repO, rw card_insert_of_not_mem,
  {rw card_filter_add_filter C wt_le, rw add_comm, apply add_le_add,
    {refl},
    {apply card_filter_wt_le_one C H}},
  {apply repO_notin}
end

def DelCodes1 {n : ℕ} : finset (finset (vector B n)) := 
  filter is_repO_in (DelCodes) 

lemma DelCodes1_def {n : ℕ} (C : finset (vector B n)) : 
  C ∈ @DelCodes1 n ↔ is_DelCode C ∧ is_repO_in C :=
by {unfold DelCodes1, rw mem_filter, rw DelCodes_def}

lemma exists_rep_O_in
  {n : ℕ} (C : finset (vector B n)) (H : is_DelCode C):
  ∃ D ∈ @DelCodes1 n, card C ≤ card D :=
begin
apply exists.intro (insert_repO C),
apply exists.intro,
  {rw DelCodes1_def, apply and.intro,
    {apply DelCode_insert_repO C H},
    {unfold is_repO_in, apply mem_insert_self}},
  {apply card_insert_repO C H}
end

def is_optimal1 {n : ℕ} (C : finset (vector B n)) := 
  is_DelCode C ∧ ∀ s ∈ @DelCodes1 n, card s ≤ card C

instance is_optimal1_decidable {n : ℕ} : 
  decidable_pred (@is_optimal1 n) :=
by {intro s, unfold is_optimal1, apply_instance}

lemma optimal_of_optimal1 
  {n : ℕ} (C : finset (vector B n)) (H : is_optimal1 C) :
  is_optimal C :=
begin
unfold is_optimal, apply and.intro,
  {apply H.left},
  {intros s sin, rw DelCodes_def at sin, 
   cases exists_rep_O_in s sin with t h_t,
   cases h_t with tin hle,
   apply le_trans hle (H.right t tin)}
end

lemma optimal1_of_optimal
  {n : ℕ} (C : finset (vector B n)) (H : is_optimal C) :
  is_optimal1 C :=
begin
unfold is_optimal1, apply and.intro,
  {apply H.left},
  {intros s sin, 
   rw DelCodes1_def at sin, cases sin with sinl sinr,
   rw ← DelCodes_def at sinl, apply H.right s sinl}
end

lemma optimal1_iff_optimal {n : ℕ} (C : finset (vector B n)) :
  is_optimal1 C ↔ is_optimal C :=
begin
apply iff.intro,
  {apply optimal_of_optimal1},
  {apply optimal1_of_optimal}
end

def DelCodes1_Area {n : ℕ} : finset (finset (vector B n)) := 
  filter is_repO_in (DelCodes_Area) 

lemma DelCodes1_Area_eq_DelCodes1 {n : ℕ} :
  @DelCodes1_Area n = @DelCodes1 n :=
begin
unfold DelCodes1_Area, unfold DelCodes1,
rw DelCodes_Area_eq_DelCodes
end

def is_optimal1_Area {n : ℕ} (C : finset (vector B n)) := 
  is_DelCode C ∧ ∀ s ∈ @DelCodes1_Area n, card s ≤ card C

instance is_optimal1_Area_decidable {n : ℕ} : 
  decidable_pred (@is_optimal1_Area n) :=
by {intro s, unfold is_optimal1_Area, apply_instance}

lemma is_optimal1_Area_iff_optimal {n : ℕ} (C : finset (vector B n)) :
  is_optimal1_Area C ↔ is_optimal C :=
begin
rw ← optimal1_iff_optimal, 
unfold is_optimal1_Area, rw DelCodes1_Area_eq_DelCodes1, 
unfold is_optimal1
end

def is_repI_in {n : ℕ} (C : finset (vector B n)) : Prop :=
  repeat I n ∈ C

instance decidable_pred_is_repI_in {n : ℕ} : 
  decidable_pred (@is_repI_in n) :=
by {intro s, unfold is_repI_in, apply_instance}

lemma repI_notin {n : ℕ} (C : finset (vector B n)):
  repeat I n ∉ filter (not wt_ge) C :=
begin
assume h, rw mem_filter at h, 
have hle : wt_ge (repeat I n), 
  {unfold wt_ge, rw wt_repI, apply nat.sub_le},
apply absurd hle h.right
end

def insert_repI {n : ℕ} (C : finset (vector B n)) : finset (vector B n) := 
  insert (repeat I n) (filter (not wt_ge) C)

lemma DelCode_insert_repI
  {n : ℕ} (C : finset (vector B n)) (H : is_DelCode C) :
  is_DelCode (insert_repI C) :=
begin
unfold insert_repI, apply DelCode_insert,
  {apply DelCode_filter C H},
  {intros y h, rw dS_inter_dS_empty_iff,
   intros i j, apply ne.symm, apply sDel_ne_repI,
   rw mem_filter at h, apply h.right}
end

lemma card_insert_repI
  {n : ℕ} (C : finset (vector B n)) (H : is_DelCode C) :
  card C ≤ card (insert_repI C) :=
begin
unfold insert_repI, rw card_insert_of_not_mem,
  {rw card_filter_add_filter C wt_ge, rw add_comm, apply add_le_add,
    {refl},
    {apply card_filter_sub_one_le_wt C H}},
  {apply repI_notin}
end

def DelCodes2 {n : ℕ} : finset (finset (vector B n)) := 
  filter is_repI_in (DelCodes) 

lemma DelCodes2_def {n : ℕ} (C : finset (vector B n)) : 
  C ∈ @DelCodes2 n ↔ is_DelCode C ∧ is_repI_in C :=
by {unfold DelCodes2, rw mem_filter, rw DelCodes_def}

lemma exists_rep_I_in
  {n : ℕ} (C : finset (vector B n)) (H : is_DelCode C):
  ∃ D ∈ @DelCodes2 n, card C ≤ card D :=
begin
apply exists.intro (insert_repI C),
apply exists.intro,
  {rw DelCodes2_def, apply and.intro,
    {apply DelCode_insert_repI C H},
    {unfold is_repI_in, apply mem_insert_self}},
  {apply card_insert_repI C H}
end

def is_optimal2 {n : ℕ} (C : finset (vector B n)) := 
  is_DelCode C ∧ ∀ s ∈ @DelCodes2 n, card s ≤ card C

instance is_optimal2_decidable {n : ℕ} : 
  decidable_pred (@is_optimal2 n) :=
by {intro s, unfold is_optimal2, apply_instance}

lemma optimal_of_optimal2 
  {n : ℕ} (C : finset (vector B n)) (H : is_optimal2 C) :
  is_optimal C :=
begin
unfold is_optimal, apply and.intro,
  {apply H.left},
  {intros s sin, rw DelCodes_def at sin, 
   cases exists_rep_I_in s sin with t h_t,
   cases h_t with tin hle,
   apply le_trans hle (H.right t tin)}
end

lemma optimal2_of_optimal
  {n : ℕ} (C : finset (vector B n)) (H : is_optimal C) :
  is_optimal2 C :=
begin
unfold is_optimal2, apply and.intro,
  {apply H.left},
  {intros s sin, 
   rw DelCodes2_def at sin, cases sin with sinl sinr,
   rw ← DelCodes_def at sinl, apply H.right s sinl}
end

lemma optimal2_iff_optimal {n : ℕ} (C : finset (vector B n)) :
  is_optimal2 C ↔ is_optimal C :=
begin
apply iff.intro,
  {apply optimal_of_optimal2},
  {apply optimal2_of_optimal}
end

def DelCodes2_Area {n : ℕ} : finset (finset (vector B n)) := 
  filter is_repI_in (DelCodes_Area) 

lemma DelCodes2_Area_eq_DelCodes2 {n : ℕ} :
  @DelCodes2_Area n = @DelCodes2 n :=
begin
unfold DelCodes2_Area, unfold DelCodes2,
rw DelCodes_Area_eq_DelCodes
end

def is_optimal2_Area {n : ℕ} (C : finset (vector B n)) := 
  is_DelCode C ∧ ∀ s ∈ @DelCodes2_Area n, card s ≤ card C

instance is_optimal2_Area_decidable {n : ℕ} : 
  decidable_pred (@is_optimal2_Area n) :=
by {intro s, unfold is_optimal2_Area, apply_instance}

lemma is_optimal2_Area_iff_optimal {n : ℕ} (C : finset (vector B n)) :
  is_optimal2_Area C ↔ is_optimal C :=
begin
rw ← optimal2_iff_optimal, 
unfold is_optimal2_Area, rw DelCodes2_Area_eq_DelCodes2, 
unfold is_optimal2
end

def DelCodes3 {n : ℕ} : finset (finset (vector B n)) := 
  filter is_repI_in DelCodes1

lemma DelCodes3_def {n : ℕ} (C : finset (vector B n)) : 
  C ∈ @DelCodes3 n ↔ is_DelCode C ∧ is_repO_in C ∧ is_repI_in C :=
by {unfold DelCodes3, rw mem_filter, rw DelCodes1_def, apply and_assoc}

lemma exists_rep_I_in'
  {n : ℕ} (H1 : 2 ≤ n) (C : finset (vector B n)) (H : is_DelCode C) :
  ∃ D ∈ @DelCodes3 n, card C ≤ card D :=
begin
apply exists.intro (insert_repI (insert_repO C)), apply exists.intro,
  {rw DelCodes3_def, apply and.intro,
    {apply DelCode_insert_repI, apply DelCode_insert_repO C H},
    {apply and.intro,
      {unfold is_repO_in, unfold insert_repI, apply mem_insert_of_mem,
       rw mem_filter, apply and.intro,
        {apply mem_insert_self},
        {unfold not, unfold wt_ge, apply not_le_of_gt, rw wt_repO, 
         apply nat.lt_of_succ_le, rw nat.le_sub_left_iff_add_le, 
          {apply H1},
          {apply nat.le_of_succ_le, apply H1}}},
      {apply mem_insert_self}}},
  {apply le_trans,
    {apply card_insert_repO C H},
    {apply card_insert_repI, apply DelCode_insert_repO C H}}
end

def is_optimal3 {n : ℕ} (C : finset (vector B n)) := 
  is_DelCode C ∧ ∀ s ∈ @DelCodes3 n, card s ≤ card C

instance is_optimal3_decidable {n : ℕ} : 
  decidable_pred (@is_optimal3 n) :=
by {intro s, unfold is_optimal3, apply_instance}

lemma optimal_of_optimal3 
  {n : ℕ} (H1 : 2 ≤ n) (C : finset (vector B n)) (H2 : is_optimal3 C) :
  is_optimal C :=
begin
unfold is_optimal, apply and.intro,
  {apply H2.left},
  {intros s sin, rw DelCodes_def at sin, 
   cases exists_rep_I_in' H1 s sin with t h_t,
   cases h_t with tin hle,
   apply le_trans hle (H2.right t tin)}
end

lemma optimal3_of_optimal
  {n : ℕ} (C : finset (vector B n)) (H : is_optimal C) :
  is_optimal3 C :=
begin
unfold is_optimal3, apply and.intro,
  {apply H.left},
  {intros s sin, 
   rw DelCodes3_def at sin, cases sin with sinl sinr,
   rw ← DelCodes_def at sinl, apply H.right s sinl}
end

lemma optimal3_iff_optimal 
  {n : ℕ} (H : 2 ≤ n) (C : finset (vector B n)) :
  is_optimal3 C ↔ is_optimal C :=
begin
apply iff.intro,
  {apply optimal_of_optimal3 H},
  {apply optimal3_of_optimal}
end
  
def DelCodes3_Area {n : ℕ} : finset (finset (vector B n)) := 
  filter is_repI_in (DelCodes1_Area) 

lemma DelCodes3_Area_eq_DelCodes3 {n : ℕ} :
  @DelCodes3_Area n = @DelCodes3 n :=
begin
unfold DelCodes3_Area, unfold DelCodes3,
rw DelCodes1_Area_eq_DelCodes1
end

def is_optimal3_Area {n : ℕ} (C : finset (vector B n)) := 
  is_DelCode C ∧ ∀ s ∈ @DelCodes3_Area n, card s ≤ card C

instance is_optimal3_Area_decidable {n : ℕ} : 
  decidable_pred (@is_optimal3_Area n) :=
by {intro s, unfold is_optimal3_Area, apply_instance}

lemma is_optimal3_Area_iff_optimal 
  {n : ℕ} (H : 2 ≤ n) (C : finset (vector B n)) :
  is_optimal3_Area C ↔ is_optimal C :=
begin
rw ← optimal3_iff_optimal H C, 
unfold is_optimal3_Area, rw DelCodes3_Area_eq_DelCodes3, 
unfold is_optimal3
end


end finset
