import .InsDel
import .NumOsNumIs
import .Moment

open list nat

namespace nat

lemma mod_congr {a b : ℕ} {m n: ℕ} (H1 : a = b) (H2 : m = n) : a % m = b % n :=
by rw H1; rw H2

lemma mod_congr_arg {a b : ℕ} (m : ℕ) (H : a = b) : a % m = b % m :=
by rw H

lemma mod_congr_func {m n : ℕ} (a : ℕ) (H : m = n)  : a % m = a % n :=
by rw H

lemma mod_neg_of_lt 
    {a b : ℕ} (H : a < b) :
    ¬ (0 < b ∧ b ≤ a) :=
begin
assume h : 0 < b ∧ b ≤ a,
apply absurd h.right,
apply not_le_of_gt H 
end 

lemma mod_neg_of_neq_zero
    {a b : ℕ} (H : ¬ a % b = 0) :
    ¬ (0 < b ∧ b ≤ b - a % b) :=
begin
assume h : 0 < b ∧ b ≤ b - a % b,  
cases b,
    {apply absurd h.left, apply not_lt_of_ge, rw ge},
    {apply absurd h.right, apply not_le_of_gt,
     apply sub_lt_of_pos_le (a % succ b) (succ b) 
      (lt_of_le_of_ne (zero_le (a % succ b)) (ne.symm H)) 
      (le_of_lt (mod_lt a (zero_lt_succ b)))}
end 

lemma mod_of_eq
    {a m : ℕ} (H : a = m) : a % m = 0 :=
begin
rw H, rw mod_self
end 

lemma mod_of_lt
    {a m : ℕ} (H : a < m) : a % m = a :=
begin
rw mod_def, rw if_neg (mod_neg_of_lt H)
end 

lemma mod_of_gt
    {a m : ℕ} (H : a > m) : a % m = (a - m) % m :=
begin
cases m,
    {rw nat.sub_zero},
    {rw mod_def, rw if_pos (and.intro (zero_lt_succ m) (le_of_lt H))},
end 

lemma mod_of_ge
    {a m : ℕ} (H : a ≥ m) : a % m =  (a - m) % m :=
begin
cases m,
    {rw nat.sub_zero},
    {rw mod_def, rw if_pos (and.intro (zero_lt_succ m) H)}
end 

def sub_mod (a b m : ℕ) : ℕ :=
if b ≤ a then (a - b) % m else (m - (b - a) % m) % m

lemma sub_mod_zero (b m : ℕ) :
    sub_mod 0 b m = (m - b % m) % m :=
begin
unfold sub_mod, cases b,
    {rw if_pos (le_refl 0), rw zero_mod, rw nat.sub_zero, rw mod_self},
    {rw if_neg (not_succ_le_zero b), rw nat.sub_zero}
end 

lemma sub_mod_sub_zero (a m : ℕ) :
    sub_mod a 0 m = a % m :=
begin
unfold sub_mod, rw if_pos (zero_le a), rw nat.sub_zero
end 

lemma sub_mod_sub_self
    (a m : ℕ) : sub_mod a a m = 0 :=
begin
unfold sub_mod, rw if_pos (le_refl a),
rw nat.sub_self, rw zero_mod
end

lemma sub_mod_mod_self (b m : ℕ) :
    sub_mod m b m = sub_mod 0 b m :=
begin
rw sub_mod, rw sub_mod_zero, 
cases decidable.em (b ≤ m) with h_le h_nle,
    {rw if_pos h_le, cases decidable.em (b = m) with h_eq h_neq,
        {rw h_eq, rw nat.sub_self, rw zero_mod,
         rw mod_self, rw nat.sub_zero, rw mod_self},
        {apply mod_congr_arg, rw mod_def, 
         rw if_neg (mod_neg_of_lt (lt_of_le_of_ne h_le h_neq))}},
    {rw if_neg h_nle, rw ← mod_of_ge (le_of_not_ge h_nle)}
end 

lemma sub_mod_add
    (a b m : ℕ) :
    sub_mod a b m = sub_mod (a + m) b m :=
begin
induction a using nat.case_strong_induction_on with a IHa,
    {rw zero_add, rw sub_mod_mod_self},
    {unfold sub_mod, cases decidable.em (b ≤ succ a) with h_le h_nle,
        {rw if_pos h_le, rw if_pos,  
            {rw add_comm, rw nat.add_sub_assoc h_le, rw add_mod_left},
            {apply le_trans h_le (le_add_right _ _)}},
        {rw if_neg h_nle, cases decidable.em (b ≤ succ a + m) with h_les h_nles,
            {rw if_pos h_les, cases decidable.em (b = succ a + m) with h_eq h_neq,
                {rw h_eq, rw nat.add_sub_cancel_left, rw mod_self,
                 rw nat.sub_zero, rw mod_self, rw nat.add_sub_add_left, 
                 rw nat.sub_self, rw zero_mod},
                {apply mod_congr_arg, rw mod_of_lt,
                    {rw add_comm, rw nat.sub_eq_iff_eq_add,
                        {rw ← nat.add_sub_assoc,
                            {rw nat.sub_add_cancel,
                                {rw nat.add_sub_cancel},
                                {rw add_comm, apply h_les}},
                            {apply le_of_lt, 
                             apply lt_of_not_ge h_nle}},
                        {apply nat.sub_le_left_of_le_add,
                         apply h_les}},
                    {apply lt_of_le_of_ne,
                        {apply nat.sub_le_left_of_le_add,
                         apply h_les},
                        {assume h, rw ← h at h_neq,
                         rw ← nat.add_sub_assoc at h_neq,
                            {rw nat.add_sub_cancel_left at h_neq,
                             contradiction},
                            {apply le_of_lt, 
                             apply lt_of_not_ge h_nle}}}}},
            {rw if_neg h_nles, rw ← nat.sub_sub, rw ← mod_eq_sub_mod,
             apply nat.le_sub_left_of_add_le (nat.le_of_lt (lt_of_not_ge h_nles))}}}
end 

lemma sub_mod_sub 
    {a m : ℕ} (b : ℕ) (H : m ≤ a) :
    sub_mod (a - m) b m = sub_mod a b m :=
begin
rw sub_mod_add (a - m) b m,
rw nat.sub_add_cancel H
end 

lemma sub_mod_mod (a b m : ℕ) :
    sub_mod a b m = sub_mod (a % m) b m  :=
begin
cases m, 
    {rw mod_zero},
    {induction a using nat.case_strong_induction_on with a IHa,
        {rw zero_mod},
        {cases decidable.em (succ a ≤ succ m) with h_le h_nle,
            {cases decidable.em (succ a = succ m) with h_eq h_neq,
                {rw h_eq, rw mod_self, rw sub_mod_mod_self},
                {rw mod_def, 
                 rw if_neg (mod_neg_of_lt (lt_of_le_of_ne h_le h_neq))}},
            {rw mod_eq_sub_mod (le_of_not_ge h_nle), rw ← IHa,
                {rw sub_mod_sub b (le_of_not_ge h_nle)},
                {rw succ_sub_succ, apply sub_le}}}}
end 

end nat

namespace list

def sub_mod (a : ℕ) (Y : list B) (m : ℕ) : ℕ :=
if (ρ Y) ≤ a then (a - ρ Y) % m else (m - ((ρ Y) - a) % m) % m

lemma sub_mod_nat.sub_mod (a m : ℕ) (Y : list B) :
    sub_mod a Y m = nat.sub_mod a (ρ Y) m :=
by trivial

lemma sub_mod_moment
    (m : ℕ) {X Y : list B} (H : (ρ Y) ≤ ρ X) :
    sub_mod (ρ X) Y m = ((ρ X) - ρ Y) % m :=
by {unfold sub_mod, rw if_pos H}

lemma sub_mod_moment_of_mod
    {a m : ℕ} {X : list B} (Y : list B) (H : (ρ X) % m = a % m) :
    sub_mod (ρ X) Y m = sub_mod a Y m :=
begin
rw sub_mod_nat.sub_mod,
rw sub_mod_nat.sub_mod,
rw nat.sub_mod_mod a,
rw nat.sub_mod_mod (ρ X),
rw H
end

lemma sub_mod_sDel
    {a m i : ℕ} {X : list B} 
    (H1 : (ρ X) % m = a % m) (H2 : length X + 1 ≤ m)  :
    sub_mod a (sDel X i) m = (ρ X) - (ρ sDel X i) :=
begin
cases X with x X',
    {rw sDel_nil, unfold sub_mod, rw moment_nil,
     rw if_pos (nat.zero_le a), rw nat.sub_zero, rw ← H1, 
     rw moment_nil, rw zero_mod},
    {rw ← sub_mod_moment_of_mod (sDel (x::X') i) H1,
     rw sub_mod_moment m (moment_sDel_le (x::X') i),
     rw nat.mod_of_lt, 
     apply lt_of_le_of_lt,
        {apply moment_sub_sDel_le, 
         apply succ_le_succ (nat.zero_le (length X'))},
        {apply lt_of_succ_le H2}}
end

lemma sub_mod_sDel_of_pos
    {a m i : ℕ} {X : list B} 
    (H1 : (ρ X) % m = a % m) (H2 : length X + 1 ≤ m)    
    (H3 : 1 ≤ length X) 
    (H4 : sub_mod a (sDel X i) m ≤ num_Is (sDel X i)) :
    sub_mod a (sDel X i) m = num_RIs X i :=
begin
rw sub_mod_sDel H1 H2,
rw sub_mod_sDel H1 H2 at H4,
apply moment_sub_sDel_of_pos H3 H4
end

lemma sub_mod_sDel_of_neg
    {a m i : ℕ} {X : list B} 
    (H1 : (ρ X) % m = a % m) (H2 : length X + 1 ≤ m)    
    (H3 : 1 ≤ length X) 
    (H4 : ¬ sub_mod a (sDel X i) m ≤ num_Is (sDel X i)) :
    sub_mod a (sDel X i) m = num_LOs X i + num_Is (sDel X i) + 1 :=
begin
rw sub_mod_sDel H1 H2,
rw sub_mod_sDel H1 H2 at H4,
apply moment_sub_sDel_of_neg H3 H4
end

lemma sIns_fig_of_pos
    {a m i : ℕ} {X : list B} 
    (H1 : (ρ X) % m = a % m) (H2 : length X + 1 ≤ m)    
    (H3 : 1 ≤ length X)
    (H4 : sub_mod a (sDel X i) m ≤ num_Is (sDel X i)) :
    X = sIns (sDel X i) i O :=
begin
rw sub_mod_sDel H1 H2 at H4,
apply sIns_fig_of_pos_of_moment H3 H4
end

lemma sIns_fig_of_neg
    {a m i : ℕ} {X : list B} 
    (H1 : (ρ X) % m = a % m) (H2 : length X + 1 ≤ m)    
    (H3 : 1 ≤ length X)
    (H4 : ¬ sub_mod a (sDel X i) m ≤ num_Is (sDel X i)) :
    X = sIns (sDel X i) i I :=
begin
rw sub_mod_sDel H1 H2 at H4,
apply sIns_fig_of_neg_of_moment H3 H4
end

end list
