import data.fintype data.vector
import .B .Moment .LevenshteinCode .DelCode .Optimality

variable {α : Type}

open vector 

def VTCode (n a: ℕ) : set (vector B n) :=
  {X| (ρ (to_list X)) % (n + 1) = a % (n + 1)}

namespace VTCode

lemma VTCode_iff (n a : ℕ) (X : vector B n) :
  VTCode n a X ↔ (ρ (to_list X)) % (n + 1) = a % (n + 1)  :=
by trivial

lemma VTCode_def (n a : ℕ) (X : vector B n) :
  X ∈ VTCode n a ↔ (ρ (to_list X)) % (n + 1) = a % (n + 1)  :=
by trivial

instance (n a : ℕ) : decidable_pred (VTCode n a) :=
λ X, decidable_of_decidable_of_iff (by apply_instance) (VTCode_iff n a X).symm

open LevenshteinCode

lemma VTCode_eq_LevCode {n a : ℕ} :
  VTCode n a = LevCode n a (n + 1) :=
begin
apply set.subset.antisymm,
  {intros X inV, rw list.def_in_LevCode, 
   apply and.intro,
    {apply inV},
    {refl}},
  {intros X inL, apply inL.left}
end

open LevenshteinCode.vector 

lemma sDelErr_correctable
  {n a : ℕ} {X ∈ VTCode n a} 
  {i : ℕ} {Y : vector B (n - 1)} (Hy : Y = sDel X i) :
  decoding_alg n a (n + 1) Y = X :=
begin
apply sDelErr_correctable Hy,
rw ← VTCode_eq_LevCode, apply H
end

open set 

theorem DelCode_VTCode {n a : ℕ} :
  is_DelCode (VTCode n a) :=
begin
intros X Y Xin Yin hneq,
apply set.subset.antisymm,
  {intros w h_w, 
   have h1 : ∃ i : ℕ, w = sDel X i, from h_w.left,
   cases h1 with i h1_i,
   have h2 : ∃ j : ℕ, w = sDel Y j, from h_w.right,
   cases h2 with j h2_j,
   have heq : X = Y,
    {rw ← @sDelErr_correctable n a X Xin i w h1_i, 
     rw ← @sDelErr_correctable n a Y Yin j w h2_j},
   apply absurd heq hneq},
  {apply set.empty_subset}
end

open finset 

def VTCode (n a : ℕ) : finset (vector B n) :=
  filter (VTCode n a) univ
  
#eval card (VTCode 10 0) 
#eval card (VTCode 10 5)
#eval card (VTCode 15 0)  
#eval card (VTCode 16 0)  
  
/-lemma card_VTCode00 : card (VTCode 0 0) = 1 := rfl
lemma card_VTCode10 : card (VTCode 1 0) = 1 := rfl
lemma card_VTCode20 : card (VTCode 2 0) = 2 := rfl
lemma card_VTCode30 : card (VTCode 3 0) = 2 := rfl
lemma card_VTCode40 : card (VTCode 4 0) = 4 := rfl
lemma card_VTCode50 : card (VTCode 5 0) = 6 := rfl
lemma card_VTCode60 : card (VTCode 6 0) = 10 := rfl
lemma card_VTCode70 : card (VTCode 7 0) = 16 := rfl-/

lemma optimal_VTCode_1_0 :
  is_optimal (VTCode 1 0) := dec_trivial

lemma optimal_VTCode_2_0 :
  is_optimal (VTCode 2 0) := dec_trivial

lemma optimal_VTCode_3_0 :
  is_optimal (VTCode 3 0) := dec_trivial

lemma optimal_VTCode_4_0 :
  is_optimal (VTCode 4 0) := dec_trivial

example : is_DelCode (VTCode 4 0) := dec_trivial

#eval @DelCodes 4

/-lemma optimal_VTCode_4_0 :
  is_optimal_Area (VTCode 4 0) := dec_trivial-/

/-lemma optimal_VTCode_4_0 :
  is_optimal1 (VTCode 4 0) := dec_trivial-/

/-lemma optimal_VTCode_4_0 :
  is_optimal1_Area (VTCode 4 0) := dec_trivial-/

/-lemma optimal_VTCode_4_0 :
  is_optimal3 (VTCode 4 0) := dec_trivial-/

/-lemma optimal_VTCode_4_0 :
  is_optimal3_Area (VTCode 4 0) := dec_trivial-/

end VTCode