import data.fintype data.vector
import .InsDel .B 

variable {α : Type}
variables [decidable_eq α]

open vector set 

def dS {n : ℕ} (X : vector α n) : set (vector α (n - 1)) :=
  {Y| ∃ i : ℕ, Y = sDel X i}

lemma dS_def {n : ℕ} (X : vector α n) (Y : vector α (n - 1)) :
  Y ∈ dS X ↔ ∃ i : ℕ, Y = sDel X i :=
by trivial

lemma sDel_ne_sDel {n : ℕ} (X Y : vector α n) 
  (H : dS X ∩ dS Y = ∅) :
  ∀ i j : ℕ, sDel X i ≠ sDel Y j :=
begin
intros i j, assume h,
have h1 : sDel X i ∈ dS X ∩ dS Y,
  {apply and.intro,
    {rw dS_def, apply exists.intro i, refl},
    {rw h, rw dS_def, apply exists.intro j, refl}},
have h2 : sDel X i ∈ ∅,
  {rw ← H, apply h1},
apply h2
end

lemma dS_inter_dS_empty {n : ℕ} (X Y : vector α n) 
  (H : ∀ i j : ℕ, sDel X i ≠ sDel Y j) :
  dS X ∩ dS Y = ∅ :=
begin
apply set.subset.antisymm,
  {intros a h, cases h with hl hr, 
   cases hl with i hl_i,
   cases hr with j hr_j,
   have h1 : sDel X i ≠ sDel Y j, 
    {apply H i j},
   have h2 : sDel X i = sDel Y j, 
    {rw ← hl_i, rw ← hr_j},
   contradiction},
  {apply set.empty_subset}
end

lemma dS_inter_dS_empty_iff {n : ℕ} (X Y : vector α n) :
  dS X ∩ dS Y = ∅ ↔ ∀ i j : ℕ, sDel X i ≠ sDel Y j :=
begin
apply iff.intro,
  {apply sDel_ne_sDel},
  {apply dS_inter_dS_empty}
end

def dS' {n : ℕ} (X : vector α n) : set (vector α (n - 1)) :=
  {Y| list.is_subseq (to_list Y) (to_list X)}

lemma dS'_def {n : ℕ} (X : vector α n) (Y : vector α (n - 1)) :
  Y ∈ dS' X ↔ list.is_subseq (to_list Y) (to_list X) :=
by trivial

instance {n : ℕ} (X : vector α n) :
  decidable_pred (dS' X) :=
λ Y, decidable_of_decidable_of_iff (by apply_instance) (dS'_def X Y).symm

lemma dS'_eq_dS {n : ℕ} (X : vector α n) :
  dS' X = dS X :=
begin
apply set.ext, 
intro y, rw dS_def, rw dS'_def, 
apply subseq_iff_exists_sDel
end

lemma dS'_iff_dS {n : ℕ} (X : vector α n) (Y : vector α (n - 1)):
  Y ∈ dS' X ↔ Y ∈ dS X :=
by rw dS'_eq_dS

instance {n : ℕ} (X : vector α n) :
  decidable_pred (dS X) :=
λ Y, decidable_of_decidable_of_iff (by apply_instance) (dS'_iff_dS X Y)


